/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <httpd.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_core.h>		/* for ap_construct_url */
#include <mod_dav.h>

#include <apr.h>
#include <apr_strings.h>
#include <apr_hash.h>
#include <apr_uuid.h>
#include <apr_file_io.h>
#include <apr_lib.h>

#include <stdio.h>
#include <mysql.h>
#include <mysqld_error.h>
#include <errmsg.h>

#include "dav_repos.h"
#include "lock.h"
#include "dbms.h"
#include "util.h"

#define DASL_MYSQL_DEBUG

/**
 * Show error message of SQL operations
 * @param db_error_message_str An application level error message string
 * @param db_errno The DB error number
 * @param dberrmessage The DB error message from the DB
 */
void static db_error_message(apr_pool_t *pool, const dav_repos_dbms *db,
	char *db_error_message_str) {
    ap_log_error(APLOG_MARK, 
		 APLOG_ERR, 0, NULL, 
		 "Error: %s\nError Number: %d\nErrorMessage:%s\n",
		 db_error_message_str, dbms_errno(db),
		 	dbms_error(pool, db));
}

/**
 * Given a particular resource (with it's serialno, type, etc) return a
 * path into the filesystem for storage. Change the algorithm to change
 * the layout of how data is stored in the fs. *path is set to NULL if
 * there is no file_dir defined (as the root dir.)
 */
static dav_error *generate_path(char **path, apr_pool_t *pool,
	const char *file_dir, const long serialno, const int version) {
    char *dirpath;

    TRACE();

    DBG2("Serial number: %ld version: %d", serialno, version);
    dirpath = apr_psprintf(pool, "%s/%ld",
    	file_dir, serialno);
    ap_no2slash(dirpath);

    // should I check to see if dirpath exists already or can I assume
    // apr_dir_make_recursive is smart 'nuf?
    if (apr_dir_make_recursive(dirpath, APR_OS_DEFAULT, pool) != APR_SUCCESS)
    	return dav_new_error(pool, HTTP_INTERNAL_SERVER_ERROR, 0,
	    "Unable to create directory for storage.");

    *path = apr_psprintf(pool, "%s/%ld-%d", dirpath, serialno, version);
    return NULL;
}

/**
 * Connects to and opens the database
 * @param d DB connection struct containing the user, password, and DB name
 * @return 0 indicating success
 */
int dbms_opendb(dav_repos_db * d)
{
    TRACE();

    dbms_create(&(d->db), d->db_id, d->db_pass, d->db_name, d->db_host, 0);
    return 0;
}

/**
 * Disconnect from the database server
 * @param d DB connection struct containing the user, password, and DB name
 * @return 0 indicating success
 */
int dbms_closedb(dav_repos_db * d)
{
    TRACE();

    dbms_destroy(&(d->db));
    return 0;
}

/**
 * r->uri for search key 
 * @param d DB connection struct containing the user, password, and DB name
 * @param r Out variable to hold the result
 * @return 1 on success, 0 for no result, -1 on fail
 */
int dbms_get_property(const dav_repos_db * d, dav_repos_resource * r)
{
    dav_repos_query *q = NULL;
    int ierrno = 0;
    int is_vhr=0, is_vr=0;
    char *uri;
    apr_pool_t *pool = r->p;

    TRACE();
    
    /* Let's check the URI is VCR or VHR */
    uri =  dav_repos_parse_vcr_uri(pool, r->uri, &(r->version), &is_vhr); 
    
    if ( uri == NULL ) /* Nor VR or VHR */ 
	uri = r->uri;
    else if (is_vhr==0)
	is_vr = 1;
    
    /* If it is VCR, you want to query version_resource using 
       uri and vnum
    */

    if (is_vr == 0) { /* It is not VCR */
        q = dbms_prepare(pool, &(d->db), "SELECT serialno,"
		"creationdate,"
		"displayname,"
		"getcontentlanguage,"
		"getcontentlength,"
		"getcontenttype,"
		"getetag,"
		"getlastmodified,"
		"resourcetype,"
		"source,"
		"depth,"
		"istext, "
		"isexternal, "
		"isversioned, "
		"checkin, "
		"checkout "
		"From dasl_resource Where URI=?");
        dbms_set_string(q, 1, uri);
    }
    else { /* It is VCR */
        q = dbms_prepare(pool, &(d->db), "SELECT vr.serialno,"
		"vr.creationdate,"
		"vr.displayname,"
		"vr.getcontentlanguage,"
		"vr.getcontentlength,"
		"vr.getcontenttype,"
		"vr.getetag,"
		"vr.getlastmodified,"
		"vr.resourcetype,"
		"vr.source,"
		"vr.depth,"
		"vr.istext, "
		"vr.isexternal "
		"From dasl_resource, version_resource vr "
		"Where dasl_resource.URI=? AND dasl_resource.serialno=vr.serialno AND "
		"vr.version=? ");
        dbms_set_string(q, 1, uri);
	dbms_set_int(q, 2, r->version);
    }
    if (dbms_execute(q)) {
        if (dbms_errno(&(d->db)) == CR_SERVER_GONE_ERROR) {
            db_error_message(r->p, &(d->db), "try reconnection");
            dbms_destroy(&(d->db));
            dbms_create(&(d->db), d->db_id, d->db_pass, d->db_name, d->db_host, 0);
                if (dbms_execute(q))  {
                    dbms_query_destroy(q); 
                    db_error_message(r->p, &(d->db), "dbms_execute error");
                    return -1;
                }
         } else {
             dbms_query_destroy(q); 
             db_error_message(r->p, &(d->db), "dbms_execute error");
             return -1;
         }
    }
    
    if ((ierrno = dbms_next(q)) < 0) {
        dbms_query_destroy(q); 
	db_error_message(r->p, &(d->db), "dbms_next error");
	return -1;
    }
    if (ierrno == 0) {
        dbms_query_destroy(q); 
    	return 0;
    }
    /*Get the data of every column of this row to a link */
    /*Fetch every column of the row */
    r->serialno = dbms_get_int(q, 1);
    /*r->filename = dbms_get_string(q, 2); For Catacomb-v*/
    r->creationdate = dbms_get_int(q, 2);
    r->displayname = dbms_get_string(q, 3);
    r->getcontentlanguage = dbms_get_string(q, 4);
    r->getcontentlength = dbms_get_int(q, 5);
    r->getcontenttype = dbms_get_string(q, 6);
    r->getetag = dbms_get_string(q, 7);
    r->getlastmodified = dbms_get_int(q, 8);
    r->resourcetype = dbms_get_int(q, 9);
    r->source = dbms_get_int(q, 10);
    r->depth = dbms_get_int(q, 11);
    r->istext = dbms_get_int(q, 12);
    r->isexternal = dbms_get_int(q, 13);

    if (is_vr == 0) {
       r->isversioned = dbms_get_int(q, 14);
       r->checkin = dbms_get_int(q, 15);
       r->checkout = dbms_get_int(q, 16);
    }

    r->next = NULL;

    dbms_query_destroy(q); 
   
    return 1;
}				/*End of dbms_get_property */

/**
 * Inserts a resource into the dasl_resource table
 * @param d DB connection struct containing the user, password, and DB name
 * @param r The resource to insert
 * @param lserialno The serial number assigned to the resource when inserted
 * @return 0 if success, -1 if fail
 */
int db_insert_resource(const dav_repos_db * d,
		       const dav_repos_resource * r, long *lserialno)
{
    dav_repos_query *q = NULL;
    int isql_result = 0;
    apr_pool_t *pool = r->p;

    TRACE();

    q = dbms_prepare(pool, &(d->db), "INSERT INTO dasl_resource \
			(uri,\
			creationdate,\
			displayname,\
			getcontentlanguage,\
			getcontentlength,\
			getcontenttype,\
			getetag,\
			getlastmodified,\
			resourcetype,\
			source,\
			depth, \
			istext, \
			isexternal ) \
			VALUES(?, ?, ?, ?, ?, ?,\
			?, ?, ?, ?, ?, ?, ?)");
    dbms_set_string(q, 1, r->uri);
    dbms_set_int(q, 2, r->creationdate);
    dbms_set_string(q, 3, apr_filepath_name_get(r->uri));
    dbms_set_string(q, 4, r->getcontentlanguage);
    dbms_set_int(q, 5, r->getcontentlength);
    dbms_set_string(q, 6, r->getcontenttype);
    dbms_set_string(q, 7, r->getetag);
    dbms_set_int(q, 8, r->getlastmodified);
    dbms_set_int(q, 9, r->resourcetype);
    dbms_set_int(q, 10, r->source);
    dbms_set_int(q, 11, r->depth);
    dbms_set_int(q, 12, r->istext);
    dbms_set_int(q, 13, r->isexternal);

    isql_result = dbms_execute(q);
    dbms_query_destroy(q); 
    if (isql_result) return isql_result;

    *lserialno = mysql_insert_id((MYSQL *) (&(d->db)));
    DBG1("SERIAL=%p", &lserialno);
    return 0;
}				/*End of db_insert_resource */

/**
 * Set a property on a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param r Identifies the resource to set the property on r->uri. 
 *          Contains the live property and value to set.
 * @return The serial number of the resource, or -1 on fail
 */

long dbms_set_property(const dav_repos_db * d, const dav_repos_resource * r)
{
    dav_repos_query *q = NULL;
    long lserialno = 0;
    int ierrno = 0;
    apr_pool_t *pool = r->p;
    int isql_result;

    TRACE();

    /*Create the select command to get the serialno by URI */
    q = dbms_prepare(pool, &(d->db), "SELECT serialno FROM dasl_resource"
			    " WHERE URI=?");
    dbms_set_string(q, 1, r->uri);

    if (dbms_execute(q)) {
	db_error_message(r->p, &(d->db), "mysql_use_result error");
	dbms_query_destroy(q); 
	return -1;
    }

    if ((ierrno = dbms_next(q)) < 0) {
	/* mysql_fetch_row() failed due to an error */
	db_error_message(r->p, &(d->db), "Failed when fetching");
	dbms_query_destroy(q); 
	return -1;
    }
    if (ierrno == 0) {
	DBG0("Create record!");
	dbms_query_destroy(q); 
	/*Insert a new resource record */
	if (db_insert_resource(d, r, &lserialno) != 0) 
	    return -1;
	else 
	    return lserialno;
    }
    lserialno = dbms_get_int(q, 1);

    dbms_query_destroy(q);

    /*Create the update command */
    q = dbms_prepare(pool, &(d->db), "UPDATE dasl_resource \
		Set creationdate=?,\
			displayname=?,\
			getcontentlanguage=?,\
			getcontentlength=?,\
			getcontenttype=?,\
			getetag=?,\
			getlastmodified=?,\
			resourcetype=?,\
			source=?,\
			depth=?,\
			istext=?, \
			isexternal=? \
			WHERE URI=?");
    dbms_set_int(q, 1, r->creationdate);
    dbms_set_string(q, 2, apr_filepath_name_get(r->uri));
    dbms_set_string(q, 3, r->getcontentlanguage);
    dbms_set_int(q, 4, r->getcontentlength);
    dbms_set_string(q, 5, r->getcontenttype);
    dbms_set_string(q, 6, r->getetag);
    dbms_set_int(q, 7, r->getlastmodified);
    dbms_set_int(q, 8, r->resourcetype);
    dbms_set_int(q, 9, r->source);
    dbms_set_int(q, 10, r->depth);
    dbms_set_int(q, 11, r->istext);
    if ( (d->file_dir != NULL) && (d->max_file_size < r->getcontentlength) ) 
    	dbms_set_int(q, 12, 1);
    else
	dbms_set_int(q, 12, 0);
    dbms_set_string(q, 13, r->uri);

    isql_result = dbms_execute(q);
    dbms_query_destroy(q); 
    if (isql_result) return -1;

    return lserialno;
}				/*End of dbms_set_property */

/**
 * Reads the text content to the buffer by serialno.
 * @param d DB connection struct containing the user, password, and DB name
 * @param bd_r Identifies the resource by bd_r->serialno
 * @return 0 on success, -1 on failure and 100 if not found
 */
int dbms_read_content(dav_repos_db * d, dav_repos_resource * db_r,
		      const char *filename)
{
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();
    
    if (d->file_dir) {
    }
    /*Create the select command to get the textcontent by serialNO */
    if (db_r->type == DAV_RESOURCE_TYPE_VERSION) { /* Version_resource */
        if (db_r->istext == 1)	/*It is a regular file */
    	    q = dbms_prepare(pool, &(d->db), "SELECT textcontent "
			"INTO DUMPFILE ? "
			"FROM version_resource "
			"WHERE serialNO=? and version = ? ");
        else if (db_r->istext == 2)	/*It is a binary file */
    	    q = dbms_prepare(pool, &(d->db), "SELECT bincontent "
			"INTO DUMPFILE ? "
			"FROM version_resource "
    			"WHERE serialNO=? and version = ? ");
        else { /* NONE */
    	    DBG0("istext is not 1 or 2.");
    	    return 100;
        }
    
        dbms_set_string(q, 1, filename);
        dbms_set_int(q, 2, db_r->serialno);
        dbms_set_int(q, 3, db_r->version);
    }
    else { /* Normal resource*/ 
        if (db_r->istext == 1)	/*It is a regular file */
    	    q = dbms_prepare(pool, &(d->db), "SELECT textcontent "
			"INTO DUMPFILE ? "
			"FROM dasl_resource "
			"WHERE serialNO=?");
        else if (db_r->istext == 2)	/*It is a binary file */
    	    q = dbms_prepare(pool, &(d->db), "SELECT bincontent "
			"INTO DUMPFILE ? "
			"FROM dasl_resource "
    			"WHERE serialNO=?");
        else { /* NONE */
    	    DBG0("istext is not 1 or 2.");
    	    return 100;
        }
    
        dbms_set_string(q, 1, filename);
        dbms_set_int(q, 2, db_r->serialno);
    }

    /*Execute the select command */
    if (dbms_execute(q)) {
	db_error_message(db_r->p, &(d->db), "mysql_query error");
	dbms_query_destroy(q); 
	return -1;
    }
    dbms_query_destroy(q);
DBG0("Before return read_content");
    return 0;
}				/*End of dbms_read_content */

/**
 * Write the text content to the dasl_resource table by serialno
 * @param d DB connection struct containing the user, password, and DB name
 * @param bd_r Identifies the resource by bd_r->serialno
 * @return 0 on success, -1 on failure
 */
int dbms_write_content(dav_repos_db * d,
		       dav_repos_resource * db_r, const char *tpfilename)
{
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();
    
    DBG1("getcontenttype: %s", db_r->getcontenttype);
    DBG1("filename: %s", tpfilename);
    /*is this a text or not? */
    if (strncasecmp(db_r->getcontenttype, "text", 4) == 0)
	q = dbms_prepare(pool, &(d->db), "update dasl_resource \
    	    set istext=1, textcontent=load_file(?) \
	    where serialno=?");
    else
	q = dbms_prepare(pool, &(d->db), "update dasl_resource \
    	    set istext=2, bincontent=load_file(?) \
	    where serialno=?");
    dbms_set_string(q, 1, tpfilename);
    dbms_set_int(q, 2, db_r->serialno);

    /*Execute the update command */
    if (dbms_execute(q)) {
	db_error_message(db_r->p, &(d->db), "Error in the update command");
	dbms_query_destroy(q); 
	return -1;
    }
    dbms_query_destroy(q); 

    return 0;
}				/*End of dbms_write_content */


/**
 * Checks a resource by its serialno whether it is in the resource link or not if exists, set locate_resource to the pointer of the resource whose serialno is 'serialno'
 * @param bd_r A linked list of resources to check
 * @param serialno The serial number of the resource to look for
 * @param resource_pointer Out variable which points to the resource, if found
 * @return 1 if the resource exists, otherwise 0 
 */
static int resource_in_link(dav_repos_resource * db_r, long serialno,
			    dav_repos_resource ** locate_resource)
{
    dav_repos_resource *resource_pointer = NULL;
    resource_pointer = db_r;

    TRACE();
    
    while (resource_pointer != NULL) {
	if (resource_pointer->serialno == serialno) {
	    *locate_resource = resource_pointer;
	    return 1;
	}

	resource_pointer = resource_pointer->next;
    }

    *locate_resource = NULL;
    return 0;
}				/*End of resource_in_link */

/**
 * Inserts a dead property struct into the dead property list of the current resource node, if name does exist in the dead property list
 * @param db_r The resource to add the property to
 * @param resource_node ???
 * @param name The name of the property to add
 * @param value The value of the property to add
 */
static void add_dead_property_inlink(dav_repos_resource * db_r,
				     dav_repos_resource * resource_node,
				     char *name, char *value)
{
    dav_repos_resource *presource_node = resource_node;
    dav_repos_property *pproperty_node = NULL;
    /*Is the previous node of pproperty_node */
    dav_repos_property *pupper_property = NULL;

    dav_repos_property *pnew_property = NULL;

    TRACE();

    pproperty_node = presource_node->pr;
    pupper_property = presource_node->pr;

    /*check whether the property with the name 'name' exists
     *in the property link or not*/

    while ((pproperty_node != NULL) &&
	   (apr_strnatcmp(pproperty_node->name, name) != 0)) {
	pupper_property = pproperty_node;
	pproperty_node = pproperty_node->next;
    }

    if (pproperty_node == NULL) {
	/*'name' does not exist, so add a new property node */
	pnew_property = apr_pcalloc(db_r->p, sizeof(*pnew_property));
	pnew_property->name = apr_pstrndup(db_r->p, name, sizeof(name));
	pnew_property->value = apr_pstrndup(db_r->p, value, sizeof(value));
	pnew_property->next = NULL;

	if (presource_node->pr == NULL)	/*Property link is originally empty */
	    presource_node->pr = pnew_property;
	else			/*Add the new node to the tail of the link */
	    pupper_property->next = pnew_property;
    }				/*End of if (pproperty_node == NULL ) */
}				/*End of add_dead_property_inlink */

/**
 * Execute the search query and return the search result in the link db_r including property link in db_r
 * @param d DB connection struct containing the user, password, and DB name
 * @param search_command SQL query to execute, the serialno field must always appear in fields of the search_command 
 * @param db_r Out variable containing the result of the search. The result starts from db_r->next
 * @return 0 on success, -1 on failure
 */
int dbms_search(dav_repos_db * d, char *search_command,
		dav_repos_resource * db_r)
{

    int isql_result = 0;
    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    int iret = 0;
    MYSQL_ROW dbrow;
    unsigned long *lengths = NULL;
    /*Pointer to the tail of the resource link */
    struct dav_repos_resource *presult_link_tail = NULL;
    /*Pointer to the new item of resource */
    struct dav_repos_resource *pnew_link_item = NULL;
    /*Pointer to resource in the link currently being set */
    struct dav_repos_resource *presource_current_set = NULL;
    int bfields_got = 0;
    unsigned int num_fields = 0;
    MYSQL_FIELD *fields = NULL;
    unsigned int i = 0;

    TRACE();

    DBG1("SEARCH: %s\n", search_command);

    /*Execute sequry */
    isql_result = mysql_query(&(d->db), search_command);
    if (isql_result != 0) {
	db_error_message(db_r->p, &(d->db), "mysql_query error");
	return -1;
    }

    use_result = mysql_use_result((MYSQL *) & (d->db));
    if (use_result == NULL) {
	db_error_message(db_r->p, &(d->db), "mysql_use_result error");
	return -1;
    }

    presult_link_tail = db_r;
    presource_current_set = db_r;
    presult_link_tail->next = NULL;
    presult_link_tail->pr = NULL;

    /*Deal with every records from the search result */
    while ((dbrow = mysql_fetch_row(use_result))) {	/*fetch every row */
	long resource_serialno = 0;	/*serialno of the resource in row */
	int resource_exists = 0;	/*resource exists in the resource link */

	if (bfields_got == 0) {	/*not got the fields yet */
	    /*Get the fields of the search command, because the fields
	     *in the command are not fixed*/
	    num_fields = mysql_num_fields(use_result);
	    if (num_fields == 0) {
		db_error_message(db_r->p, &(d->db), "Error in mysql_num_fields.");
		mysql_free_result(use_result);
		return -1;
	    }
	    fields = mysql_fetch_fields(use_result);
	    if (fields == NULL) {
		db_error_message(db_r->p, &(d->db), "Error in mysql_fetch_fields.");
		mysql_free_result(use_result);
		return -1;
	    }
	    bfields_got = 1;
	}			/*End of if ( bfields_got == 0) */

	/*Get the length of every field */
	lengths = mysql_fetch_lengths(use_result);
	if (lengths == NULL) {
	    db_error_message(db_r->p, &(d->db), "Error in mysql_fetch_lengths.");
	    mysql_free_result(use_result);
	    return -1;
	}

	/*Check where resource has existed in the resource link */
	resource_serialno = atol(dbrow[0]);
	resource_exists =
	    resource_in_link(db_r->next, resource_serialno,
			     &presource_current_set);

	if (resource_exists == 0) {	/*Resource not in the resource link */
	    /*Resource not the resource link then create a new struct node */
	    /* Malloc a new structure */
	    
	    pnew_link_item = dav_repos_mk_blank_resource(db_r->p);
	    presult_link_tail->next = pnew_link_item;
	    /* Copy pool */
	    pnew_link_item->p = db_r->p;
	    presult_link_tail = pnew_link_item;

	    /*Set value of the record by every field */
	    for (i = 0; i < num_fields; i++) {
	        /* # FIXME
		   fields[i] or fields[i].name is NULL 
		*/
	        if (fields[i].name == NULL)
		  continue;

		/*i is the number of the column starting from 0 */
		/*Get the data of every column of this row to a link */
		if (apr_strnatcasecmp(fields[i].name, "serialno") == 0)
		    presult_link_tail->serialno = atol(dbrow[i]);

		else if (apr_strnatcasecmp(fields[i].name, "URI") == 0)
		    presult_link_tail->uri =
			(dbrow[i] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[i], lengths[i]);
		else if (apr_strnatcasecmp(fields[i].name, "creationdate") == 0)
		    presult_link_tail->creationdate = atoll(dbrow[i]);
		else if (apr_strnatcasecmp(fields[i].name, "displayname") == 0)
		    presult_link_tail->displayname =
			(dbrow[i] == NULL) ? NULL : apr_pstrndup(db_r->p,
								 dbrow[i],
								 lengths[i]);
		else if (apr_strnatcasecmp
			 (fields[i].name, "getcontentlanguage") == 0)
		    presult_link_tail->getcontentlanguage =
			(dbrow[i] == NULL) ? NULL : apr_pstrndup(db_r->p,
								 dbrow[i],
								 lengths[i]);
		else if (apr_strnatcasecmp(fields[i].name, "getcontenttype")
			 == 0)
		    presult_link_tail->getcontenttype =
			(dbrow[i] == NULL) ? NULL : apr_pstrndup(db_r->p,
								 dbrow[i],
								 lengths[i]);
		else if (apr_strnatcasecmp(fields[i].name, "getcontentlength")
			 == 0)
		    presult_link_tail->getcontentlength = atol(dbrow[i]);
		else if (apr_strnatcasecmp(fields[i].name, "getetag") == 0)
		    presult_link_tail->getetag =
			(dbrow[i] == NULL) ? NULL : apr_pstrndup(db_r->p,
								 dbrow[i],
								 lengths[i]);
		else if (apr_strnatcasecmp(fields[i].name, "getlastmodified")
			 == 0)
		    presult_link_tail->getlastmodified = atoll(dbrow[i]);
		
		else if (apr_strnatcasecmp(fields[i].name, "resourcetype") ==
			 0)
		    presult_link_tail->resourcetype = atoi(dbrow[i]);
		else if (apr_strnatcasecmp(fields[i].name, "source") == 0)
		    presult_link_tail->source = atoi(dbrow[i]);
		
		else if (apr_strnatcasecmp(fields[i].name, "depth") == 0)
		    presult_link_tail->depth = atoi(dbrow[i]);
		else if (apr_strnatcasecmp(fields[i].name, "istext") == 0)
		    presult_link_tail->istext = atoi(dbrow[i]);
	    }			/*End of for every field */

	    /*Used when setting dead properties */
	    presource_current_set = presult_link_tail;
	}			/*End of if resource_in_link == 0 */

	/*deal with the dead property section in this row */
	i = 0;

	/*Locate the dead properties, 'name' and 'value', 
	   in the query command */
	while ((i < num_fields) &&
	       (apr_strnatcasecmp(fields[i].name, "name") != 0))
	    i++;

	if (i < num_fields) {	/*Found the name field */
	    /*Insert a dead property node to the dead property link
	     * in the current resource node, if name does exist in the
	     * dead property link*/
	    add_dead_property_inlink(db_r, presource_current_set,
				     dbrow[i] /*name */ ,
				     dbrow[i + 1] /*value */ );
	    if (isql_result != 0) {
		mysql_free_result(use_result);
		return -1;
	    }
	}

	presult_link_tail->next = NULL;
	presult_link_tail->pr = NULL;
    }				/*End of while (( dbrow = ... )) */

    ierrno = mysql_errno((MYSQL *) & (d->db));
    if (ierrno != 0) {		/* mysql_fetch_row() failed due to an error */
	db_error_message(db_r->p, &(d->db), "Failed when fetching");
	iret = -1;
    }
    else
	iret = 0;

    mysql_free_result(use_result);

    presult_link_tail = db_r;
    while (presult_link_tail != NULL) {
	presult_link_tail = presult_link_tail->next;
    }

    return iret;
}				/*End of dbms_search */


/**
 * Removes a resource from the repository. If the resource is collection, all the resources it contains will also be removed.
 * @param d DB connection struct containing the user, password, and DB name.
 * @param r A pointer to the resource to be removed.
 * @return 0 on success, -1 on failure
 */
int dbms_remove_resource(const dav_repos_db * d, const dav_repos_resource * r)
{
    apr_pool_t *pool = r->p;
    dav_repos_query *q = NULL;
    char *path = NULL;

    TRACE();

    /* You can't delete it directly */
    if (r->type == DAV_RESOURCE_TYPE_VERSION)
	return -1;

    if (r->type == DAV_RESOURCE_TYPE_HISTORY) { /* Version_histpry */
        q = dbms_prepare(pool, &(d->db), "DELETE FROM version_resource "
			 "WHERE serialno=?");
        dbms_set_int(q, 1, r->serialno);
    }
    else { /* Normal resource */
        if (r->resourcetype == 0) {	/*Is a regular file, not directory */
    	    q = dbms_prepare(pool, &(d->db), "DELETE FROM dasl_resource \
			       WHERE serialno=?");
            dbms_set_int(q, 1, r->serialno);
            /* delete the file-content on harddisk if it is external resource */
            if ((r->isexternal==1) && (!generate_path (&path, r->p, d->file_dir, r->serialno, r->version))) {
                apr_file_remove(path, r->p);
                char *deldir=NULL;
                deldir = apr_psprintf(pool, "%s/%ld",d->file_dir, r->serialno);
                DBG1("delete directory %s", deldir);    
                apr_dir_remove(deldir, r->p);
            }
        }				/*End of if it is a regular file */
        else if (r->resourcetype == 1) {	/*It is a collection(directory) */
            /* first delete harddisk files */
            q = dbms_prepare(pool, &(d->db), "SELECT serialno \
                                                  FROM dasl_resource \
                                                  WHERE URI like ? \
                                                  AND isexternal=1");
            dbms_set_string(q, 1, apr_psprintf(pool, "%s%%", r->uri));
            if (dbms_execute(q)) {
                db_error_message(pool, &(d->db),
                        "Error in the select command (dbms_remove_resource)");
                dbms_query_destroy(q);
                return -1;
            }
            while( dbms_next(q) > 0 ) {
                generate_path (&path, pool, d->file_dir, dbms_get_int(q, 1), r->version);
                apr_file_remove(path, pool);
                char *deldir=NULL;
                deldir = apr_psprintf(pool, "%s/%ld",d->file_dir, (long)dbms_get_int(q, 1));
                DBG1("delete directory %s", deldir);
                apr_dir_remove(deldir, r->p);
            }
            dbms_query_destroy(q);

            /* now delete database content */
            q = dbms_prepare(pool, &(d->db), "DELETE FROM dasl_resource \
                WHERE URI=? or URI like ? '/%'");
            dbms_set_string(q, 1, r->uri);
            dbms_set_string(q, 2, r->uri);
        } else { /* huh? */
     	    return -1;
        }
    }

    /*Execute the delete command */
    if (dbms_execute(q)) {
	dbms_query_destroy(q); 
	db_error_message(r->p, &(d->db), "Error in the delete command");
	return -1;
    }
    dbms_query_destroy(q); 

    /* Set the version things in the 'dasl_resource' */
    if (r->type == DAV_RESOURCE_TYPE_HISTORY) { /* Version_property */
        q = dbms_prepare(pool, &(d->db), "UPDATE dasl_resource "
		                 	 "SET isversioned = 0, checkin=-1, checkout = -1 "
					 "WHERE serialno=?");
        dbms_set_int(q, 1, r->serialno);
        if (dbms_execute(q)) {
	    dbms_query_destroy(q); 
	    db_error_message(r->p, &(d->db), "Error in the update command");
	    return -1;
	}
	
        dbms_query_destroy(q); 
    }

    return 0;
}				/*End of dbms_remove_resource */

/**
 * Adds a new dead property record the the dead property table in the DB
 * @param d DB connection struct
 * @param r Identifies the resource to associate the property with
 * @param pr The property to insert
 * @return 0 on success, -1 on failure
 */
int db_insert_property(const dav_repos_db * d,
		       const dav_repos_resource * r,
		       const dav_repos_property * pr)
{
    apr_pool_t *pool = r->p;
    dav_repos_query *q = NULL;

    TRACE();

    /*Create the insert command */
    q = dbms_prepare(pool, &(d->db), "INSERT INTO dasl_property "
			    "(serialno, ns_id, name, value) "
			    "VALUES(?, ?, ?, ?)");
    dbms_set_int(q, 1, r->serialno);
    dbms_set_int(q, 2, pr->ns_id);
    dbms_set_string(q, 3, pr->name);
    dbms_set_string(q, 4, pr->value);

    if (dbms_execute(q)) {
        dbms_query_destroy(q); 
	db_error_message(r->p, &(d->db), "Error in the update command");
	return -1;
    }
    dbms_query_destroy(q); 

    return 0;
}				/*End of db_insert_property */

/**
 * Set the value of a dead property on a resource identified by serialno
 * @param d DB connection struct
 * @param r Identified the resource to set the property on
 * @param pr The dead property to set
 * @return 0 on success, -1 on failure
 */
int dbms_set_dead_property(const dav_repos_db * d,
			   const dav_repos_resource * r,
			   const dav_repos_property * pr)
{
    int ierrno = 0;
    apr_pool_t *pool = r->p;
    dav_repos_query *q = NULL;

    TRACE();
	
    /*Create the select command to get the serialno by serialno */
    q = dbms_prepare(pool, &(d->db), "SELECT serialno FROM dasl_property"
			    " WHERE serialno=? and name=? and ns_id=?");
    dbms_set_int(q, 1, r->serialno);
    dbms_set_string(q, 2, pr->name);
    dbms_set_int(q, 3, pr->ns_id);

    if (dbms_execute(q)) {
        dbms_query_destroy(q); 
	db_error_message(r->p, &(d->db), "mysql_use_result error");
	return -1;
    }

    ierrno = dbms_next(q);
    dbms_query_destroy(q);
    if (ierrno == 0) {
	   /*Insert a new property record */
	   return db_insert_property(d, r, pr);
    } 
    else if (ierrno == -1) {
	   db_error_message(r->p, &(d->db), "Failed when fetching");
	   return -1;
    }

    /*Create update command */
    q = dbms_prepare(pool, &(d->db), "UPDATE dasl_property \
		       		set ns_id=?, name=?, value=? \
			       WHERE serialno=? and name=?");
    dbms_set_int(q, 1, pr->ns_id);
    dbms_set_string(q, 2, pr->name);
    dbms_set_string(q, 3, pr->value);
    dbms_set_int(q, 4, r->serialno);
    dbms_set_string(q, 5, pr->name);

    /*Execute the update command */
    if (dbms_execute(q)) {
        dbms_query_destroy(q);
	    db_error_message(r->p, &(d->db), "Error in the update command");
	   return -1;
    }
    dbms_query_destroy(q); 

    return 0;
}				/*End of dbms_set_dead_property */


/**
 *	I'll fill pr->name , r->serialno and you will fill pr->value 
 * @param d DB connection struct
 * @param r Identifies the resource
 * @param pr Identifies the property to retrieve
 * @return 0 on success, 100 if not found, -1 on failure
 * */
int dbms_get_dead_property(const dav_repos_db * d,
			   const dav_repos_resource * r,
			   dav_repos_property * pr)
{
    int iret = 0;
    apr_pool_t *pool = r->p;
    dav_repos_query *q = NULL;
    char *value;

    TRACE();

    /*Create the search command */
    if (r->type == DAV_RESOURCE_TYPE_VERSION) { /* Version_property */
        q = dbms_prepare(pool, &(d->db), "SELECT ns_id, value FROM version_property "
			"Where serialno=? and version=? and name=? and ns_id=?");
        dbms_set_int(q, 1, r->serialno);
        dbms_set_int(q, 2, r->version);
        dbms_set_string(q, 3, pr->name);
        dbms_set_int(q, 4, pr->ns_id);
    }
    else { /* Normal property */
        q = dbms_prepare(pool, &(d->db), "SELECT ns_id, value FROM dasl_property "
			"Where serialno=? and name=? and ns_id=?");
        dbms_set_int(q, 1, r->serialno);
        dbms_set_string(q, 2, pr->name);
        dbms_set_int(q, 3, pr->ns_id);
    }
    
    if (dbms_execute(q)) {
	dbms_query_destroy(q);
	db_error_message(r->p, &(d->db), "mysql_use_result error");
	return -1;
    }

    if ((iret = dbms_next(q)) == 1) {
    	pr->ns_id = dbms_get_int(q, 1);
	value = dbms_get_string(q, 2);
	pr->next = NULL;
    }

    dbms_query_destroy(q); 

    return iret;
}				/*End of dbms_dead_property */

/**
 * Moves a resource. If the source is a regular file, 
 * just change the URI. If the source is collection, 
 * move all files in the collection 
 * @param d DB connection struct
 * @param r_src source 
 * @param r_des destination ( only contains URI field )
 * @return 0 on success -1 on failure
 */
int dbms_move_resource(const dav_repos_db * d,
		       const dav_repos_resource * r_src,
		       const dav_repos_resource * r_des)
{
    int des_depth = 0;		/* Depth of the destination URI */
    int depth_diff = 0;
    apr_pool_t *pool = r_src->p;
    dav_repos_query *q = NULL;

    TRACE();

    des_depth = ap_count_dirs(r_des->uri);
    depth_diff = des_depth - ap_count_dirs(r_src->uri);

    if (r_src->resourcetype == 0) {
	/*Is a regular file, not directory
	 *Change the URI directly*/
	q = dbms_prepare(pool, &(d->db), "UPDATE dasl_resource \
				SET URI=?, \
                displayname=?, \
				depth=? \
				WHERE serialno=?");
	dbms_set_string(q, 1, r_des->uri);
	dbms_set_string(q, 2, apr_pstrdup(pool, apr_filepath_name_get(r_des->uri)) );
	dbms_set_int(q, 3, des_depth);
	dbms_set_int(q, 4, r_src->serialno);
    }				/*End of if it is a regular file */
    else if (r_src->resourcetype == 1) {  /*It is a collection(directory) */
    	q = dbms_prepare(pool, &(d->db),
				"UPDATE dasl_resource "
				"SET URI = concat(?, SUBSTRING(URI,?)), "
				"depth=depth + ? "
				"WHERE URI = ? or URI like ? '/%'");
	dbms_set_string(q, 1, r_des->uri);
	dbms_set_int(q, 2, strlen(r_src->uri) + 1);
	dbms_set_int(q, 3, depth_diff);
	dbms_set_string(q, 4, r_src->uri);
	dbms_set_string(q, 5, r_src->uri);
    } else { /* huh? */
      	return -1;
    }

    if (dbms_execute(q)) {
	db_error_message(r_src->p, &(d->db), "Error in the create temporary command");
	dbms_query_destroy(q); 
	return -1;
    }

    dbms_query_destroy(q);

    return 0;
}				/*End of dbms_move_resource */

/**
 * Copy the source to the destination. If the source is a regular file, 
 * just copy the record in the database with the new URI. 
 * If the source is collection, duplicate all records in the collection 
 * in the database with the new URIs
 * @param d DB connection struct
 * @param r_src The source
 * @param r_des The destination ( only contains URI field )
 * @return 0 on success, -1 on failure
 */
int dbms_copy_resource(const dav_repos_db * d,
		       const dav_repos_resource * r_src,
		       const dav_repos_resource * r_des,
		       int depth)
{
    int des_depth = 0;		/* Depth of the destination URI */
    int depth_diff = 0;
    apr_pool_t *pool = r_src->p;
    dav_repos_query *q = NULL;
    char *src_fn = NULL, *des_fn = NULL;

    TRACE();

    des_depth = ap_count_dirs(r_des->uri);
    depth_diff = des_depth - ap_count_dirs(r_src->uri);

    /* Drop the temporary table first */
    q = dbms_prepare(pool, &(d->db), "drop table tmp_resrecord");
    if (dbms_execute(q)) {
	DBG0("Error in the drop table command ");
    }
    dbms_query_destroy(q);

    if (r_src->resourcetype == 0 || depth == 0) {
	/*Source Is a regular file, not directory */
	/*Copy the record in the database with the new URI */
	/*Store the source record into a temporary table */
	q = dbms_prepare(pool, &(d->db),
				"CREATE temporary table tmp_resrecord "
				"SELECT 0, "
				"? AS uri,"
				"creationdate,"
				"displayname,"
				"getcontentlanguage,"
				"getcontentlength,"
				"getcontenttype,"
				"getetag,"
				"getlastmodified,"
			        "resourcetype,"
				"source,"
				"?,"
				"istext, "
				"isexternal, "
				"textcontent, "
				"bincontent, "
				"0+0, "
				"-1, "
				"-1+0 "
				"FROM dasl_resource "
				"WHERE serialno=?");
	dbms_set_string(q, 1, r_des->uri);
	dbms_set_int(q, 2, des_depth);
	dbms_set_int(q, 3, r_src->serialno);

	if (dbms_execute(q)) {
	    db_error_message(r_src->p, &(d->db),
			     "Error in the create temporary command");
            dbms_query_destroy(q); 
	    return -1;
	}
	dbms_query_destroy(q);
	/* Duplicate the records in the temporary table 
	   to the table dasl_resource */
	q = dbms_prepare(pool, &(d->db), "INSERT into dasl_resource "
		"SELECT * from tmp_resrecord");
	if (dbms_execute(q)) {
	    db_error_message(r_src->p, &(d->db), "Error in the insert command");
            dbms_query_destroy(q); 
	    return -1;
	}
    dbms_query_destroy(q);
    
    // now copy the dead properties
    dbms_copy_dead_properties(d, r_src, r_des);
	
	if (r_src->isexternal && r_src->resourcetype==0) {
            long des_serialno = mysql_insert_id((MYSQL *) (&(d->db)));
	    DBG0("file is externally stored, copying...");
	    if (generate_path (&src_fn, r_src->p, d->file_dir,
		    r_src->serialno, r_src->version)) {
		
		db_error_message(r_src->p, &(d->db),
		    "Error generating source path.");
		return -1;
	    }
	    // new resource is version 0 by default
	    if (generate_path (&des_fn, r_src->p, d->file_dir,
		    des_serialno, 0)) {
		db_error_message(r_src->p, &(d->db),
		    "Error generating destination path.");
		return -1;
	    }
	    DBG2("copying file %s to %s", src_fn, des_fn);
	    if (apr_file_copy(src_fn, des_fn, APR_OS_DEFAULT, r_src->p)
		    != APR_SUCCESS) {
		db_error_message(r_src->p, &(d->db), "Unable to copy file");
		return -1;
	    }
	}
    }				/*End of if it is a regular file */
    else if (r_src->resourcetype == 1) {
	/*It is a collection(directory) */
	/* duplicate the records in the collection 
	   in the database with the new URIs */
	/*Store the source records into a temporary table */
	q = dbms_prepare(pool, &(d->db),
				"CREATE temporary table tmp_resrecord "
				"SELECT 0, "
				"concat(?, SUBSTRING(URI,?)) AS x, "
				"creationdate,"
				"displayname,"
				"getcontentlanguage,"
				"getcontentlength,"
				"getcontenttype,"
				"getetag,"
				"getlastmodified,"
				"resourcetype,"
				"source,"
				"depth + ?, "
				"istext, "
				"isexternal, "
				"textcontent, "
				"bincontent, "
				"0+0, "
				"-1, "
				"-1+0 "
				"FROM dasl_resource "
				"WHERE URI=? or URI like ? '/%'");
	dbms_set_string(q, 1, r_des->uri);
	dbms_set_int(q, 2, strlen(r_src->uri) + 1);
	dbms_set_int(q, 3, depth_diff);
	dbms_set_string(q, 4, r_src->uri);
	dbms_set_string(q, 5, r_src->uri);

	/*Execute the command */
	if (dbms_execute(q)) {
	    DBG0("Error in the create temporary command");
        dbms_query_destroy(q);
	    return -1;
	}
	dbms_query_destroy(q);

	/* Duplicate the records in the temporary table 
	   to the table dasl_resource */
	q = dbms_prepare(pool, &(d->db), "INSERT into dasl_resource "
		"SELECT * from tmp_resrecord");
	if (dbms_execute(q)) {
	    DBG0("Error in the insert command");
        dbms_query_destroy(q); 
	    return -1;
	}
	dbms_query_destroy(q);

    // copy the dead properties
    dbms_copy_dead_properties(d, r_src, r_des);

	q = dbms_prepare(pool, &(d->db), "select s.serialno, d.serialno from dasl_resource s, dasl_resource d where s.uri like ? and d.uri like ? and s.resourcetype = 0 and d.resourcetype = 0 and s.isexternal != 0 and d.isexternal != 0 and substring(s.uri, ?) = substring(d.uri, ?)");
	dbms_set_string(q, 1, apr_psprintf(r_src->p, "%s/%%", r_src->uri));
	dbms_set_string(q, 2, apr_psprintf(r_src->p, "%s/%%", r_des->uri));
	dbms_set_int(q, 3, strlen(r_src->uri) + 1);
	dbms_set_int(q, 4, strlen(r_des->uri) + 1);
	if (!(dbms_select(q))) {
	   DBG0("Error retrieving dest files");
        dbms_query_destroy(q);
	    return -1;
	}
        while (dbms_next(q)) {
	    char *src_fn = NULL, *des_fn = NULL;
	    long src_serialno = dbms_get_int(q, 1);
	    long des_serialno = dbms_get_int(q, 2);

	    if (generate_path (&src_fn, r_src->p, d->file_dir,
		    src_serialno, 0)) {
		db_error_message(r_src->p, &(d->db),
		    "Error generating source path.");
		dbms_query_destroy(q); 
		return -1;
	    }

	    if (generate_path (&des_fn, r_src->p, d->file_dir,
		    des_serialno, 0)) {
		db_error_message(r_src->p, &(d->db),
		    "Error generating destination path.");
		dbms_query_destroy(q); 
		return -1;
	    }

	    if (apr_file_copy(src_fn, des_fn, APR_OS_DEFAULT, r_src->p)
		    != APR_SUCCESS) {
		db_error_message(r_src->p, &(d->db),
		    "Unable to copy file");
		dbms_query_destroy(q); 
		return -1;
	    }
	}
        dbms_query_destroy(q); 
    }				/*End of else if */

    return 0;
}				/*End of dbms_copy_resource */


/**
 * Copy all dead properties from the source to the destination resource. 
 * If the source is a regular file, just copy properties of one resource. 
 * If the source is collection, all properties  will be recursivly copied. 
 * @param d DB connection struct
 * @param r_src The source 
 * @param r_des The destination ( only contains URI field )
 * @return 0 on success, -1 on failure
 */
int dbms_copy_dead_properties(const dav_repos_db * d,
                  const dav_repos_resource * r_src,
                  const dav_repos_resource * r_des)
{
    apr_pool_t *pool = r_src->p;
    dav_repos_query *q = NULL;
    int i=0, ierrno=0, numcopy=0;

    TRACE();

    // copy dead properties of all resources
    q = dbms_prepare(r_src->p, &(d->db), "SELECT COUNT(s.serialno) \
                                        FROM dasl_resource s, dasl_resource d \
                                        WHERE s.uri like ? and d.uri like ? \
                                        AND substring(s.uri, ?) = substring(d.uri, ?)");
    dbms_set_string(q, 1, apr_psprintf(r_src->p, "%s/%%", r_src->uri));
    dbms_set_string(q, 2, apr_psprintf(r_src->p, "%s/%%", r_des->uri));
    dbms_set_int(q, 3, strlen(r_src->uri) + 1);
    dbms_set_int(q, 4, strlen(r_des->uri) + 1);

    if (dbms_execute(q)) {
        db_error_message(r_src->p, &(d->db),
                "Error in the count dead properties command");
        dbms_query_destroy(q);
        return -1;
        }
    numcopy = dbms_get_int(q, 1);
    dbms_query_destroy(q); 

    int *src_ptr = apr_pcalloc(pool, 2 * (numcopy+1) * sizeof(int));
    int *des_ptr = apr_pcalloc(pool, 2 * (numcopy+1) * sizeof(int));
    q = dbms_prepare(r_src->p, &(d->db), "SELECT s.serialno, \
                                        d.serialno \
                                        FROM dasl_resource s, dasl_resource d \
                                        WHERE s.uri like ? and d.uri like ? \
                                        AND substring(s.uri, ?) = substring(d.uri, ?)");
    dbms_set_string(q, 1, apr_psprintf(r_src->p, "%s%%", r_src->uri));
    dbms_set_string(q, 2, apr_psprintf(r_src->p, "%s%%", r_des->uri));
    dbms_set_int(q, 3, strlen(r_src->uri) + 1);
    dbms_set_int(q, 4, strlen(r_des->uri) + 1);
    dbms_execute(q);
    while ((ierrno = dbms_next(q)) > 0) {
        src_ptr[i] = dbms_get_int(q, 1);
        des_ptr[i] = dbms_get_int(q, 2); 
        i++;
    }
    dbms_query_destroy(q);

    for(i=0; i<=numcopy;i++) { 

        /* Drop the temporary table */
        q = dbms_prepare(pool, &(d->db), "drop table tmp_resprops");
        if (dbms_execute(q)) {
            DBG0("Error in the drop table command ");
        }
        dbms_query_destroy(q);

        q = dbms_prepare(pool, &(d->db),
                    "CREATE temporary table tmp_resprops "
                    "SELECT ?, "
                    "ns_id, "
                    "name, "
                    "value "
                    "FROM dasl_property "
                    "WHERE serialno = ?");
        dbms_set_int(q, 1, des_ptr[i]);
        dbms_set_int(q, 2, src_ptr[i]);
        /*Execute the command */
        if (dbms_execute(q)) {
        db_error_message(r_src->p, &(d->db),
                "Error in the create temporary command");
        dbms_query_destroy(q);
        return -1;
        }
        dbms_query_destroy(q);

        q = dbms_prepare(pool, &(d->db), "INSERT into dasl_property "
            "SELECT * from tmp_resprops");
        if (dbms_execute(q)) {
            db_error_message(r_src->p, &(d->db), "Error in the insert command");
            dbms_query_destroy(q); 
            return -1;
        }
        dbms_query_destroy(q);
    }
  return(0);
}

/**
 * Exists check using URI
 * @param d DB connection struct
 * @param r A pointer to a resource struct containing the memory pool to use
 * @param uri The URI of the resource to check
 * Return 0: Not Found 1: Found -1: Failure
 */
int dbms_exists(const dav_repos_db * d, const char *uri, apr_pool_t *pool)
{
    int ierrno = 0;
    dav_repos_query *q = NULL;

    TRACE();

    /*Create the search command */
    q = dbms_prepare(pool, &(d->db), "SELECT serialno "
				  "From dasl_resource "
				  "WHERE uri=?");
    dbms_set_string(q, 1, uri);

    if (dbms_execute(q)) {
    	dbms_query_destroy(q);
	return -1;
    }
    
    ierrno = dbms_next(q);
    dbms_query_destroy(q);
    return ierrno;
}				/*End of dbms_exists */


/**
 * Get properties in a form of property link for every resource
 * @param d DB connection struct
 * @param db_r
 * @return 0 on success, -1 on failure
 */
int dbms_fill_dead_property(const dav_repos_db * d, dav_repos_resource * db_r)
{    
    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    MYSQL_ROW dbrow;
    unsigned long *lengths = NULL;
    struct dav_repos_resource *presource_link_tail = NULL;
    struct dav_repos_property *pproperty_link_tail = NULL;
    struct dav_repos_property *pnew_link_item = NULL;
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();

    presource_link_tail = db_r;

    while (presource_link_tail != NULL) {
	/*Create the search command */
	q = dbms_prepare(pool, &(d->db), "SELECT serialno,\
                        ns_id, \
			name,\
			value \
			From dasl_property Where serialno=? \
			ORDER BY name");
	dbms_set_int(q, 1, presource_link_tail->serialno);

	if (!(use_result = dbms_select(q))) {
	    db_error_message(db_r->p, &(d->db), "mysql_use_result error");
	    dbms_query_destroy(q); 
	    return -1;
	}

	while ((dbrow = mysql_fetch_row(use_result))) {
	    /*fetch every property for a resource */
	    /*Get the length of every field */
	    lengths = mysql_fetch_lengths(use_result);
	    if (lengths == NULL) {
		db_error_message(db_r->p, &(d->db),
				 "Error in mysql_fetch_lengths.");
		dbms_query_destroy(q); 
		return -1;
	    }

	    /*Malloc a new property structure */
	    pnew_link_item = apr_pcalloc(db_r->p, sizeof(*pnew_link_item));

	    if (presource_link_tail->pr == NULL) {
		/*It is the first property record */
		presource_link_tail->pr = pnew_link_item;
		pproperty_link_tail = pnew_link_item;
	    }
	    else {
		/*It is not the first property record */
		if (pproperty_link_tail)
		    pproperty_link_tail->next = pnew_link_item;
		pproperty_link_tail = pnew_link_item;
	    }

	    /*Get the data of every column of this row to a link */

	    pproperty_link_tail->serialno = atol(dbrow[0]);

	    pproperty_link_tail->ns_id = atol(dbrow[1]);

	    pproperty_link_tail->name =
		(dbrow[2] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[2],
							 lengths[2]);

	    pproperty_link_tail->value =
		(dbrow[3] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[3],
							 lengths[3]);

	    pproperty_link_tail->next = NULL;
	}			/*End of while (( dbrow = ... )) */

	ierrno = mysql_errno((MYSQL *) & (d->db));
	if (ierrno != 0) {	/* mysql_fetch_row() failed due to an error */
	    db_error_message(db_r->p, &(d->db), "Failed when fetching");
	    dbms_query_destroy(q); 
	    return -1;
	}

        dbms_query_destroy(q); 
 
	presource_link_tail = presource_link_tail->next;	/*Next resource */
    }				/*End of while presource_link_tail != NULL */

    return 0;
}				/*End of db_get_collection_property */

/**
 * Retrieve all the resources in a collection 
 * @param d DB connection struct
 * @param db_r A pointer to a resource struct which identifies the collection
 * @param depth==1 one depth, other infinity
 * @return 0 on success, -1 on failure
 */
int dbms_get_collection_resource(const dav_repos_db * d,
				 dav_repos_resource * db_r, int depth)
{
    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    int iret = 0;
    MYSQL_ROW dbrow;
    unsigned long *lengths = NULL;
    struct dav_repos_resource *presult_link_tail = NULL;
    struct dav_repos_resource *pnew_link_item = NULL;
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();

    /*Create the search command */
    q = dbms_prepare(pool, &(d->db), "SELECT serialno,\
		URI,\
		creationdate,\
		displayname,\
		getcontentlanguage,\
		getcontentlength,\
		getcontenttype,\
		getetag,\
		getlastmodified,\
		resourcetype,\
		source,\
		depth, \
		istext, \
		isversioned, \
		checkin, \
		checkout, \
		isexternal \
		From dasl_resource Where URI like ? '/%' \
                and URI <> ? \
		and (? <> 1 or depth = ?) \
		ORDER by resourcetype DESC, uri, serialno  ");
    dbms_set_string(q, 1, db_r->uri);
    dbms_set_string(q, 2, db_r->uri);
    dbms_set_int(q, 3, depth);
    dbms_set_int(q, 4, db_r->depth + 1);

    if (!(use_result = dbms_select(q))) {
	db_error_message(db_r->p, &(d->db), "mysql_use_result error");
	dbms_query_destroy(q);
	return -1;
    }

    presult_link_tail = db_r;
    presult_link_tail->next = NULL;
    presult_link_tail->pr = NULL;
    while ((dbrow = mysql_fetch_row(use_result))) {
	/*fetch every row */
	/*Get the length of every field */
	lengths = mysql_fetch_lengths(use_result);
	if (lengths == NULL) {
	    db_error_message(db_r->p, &(d->db), "Error in mysql_fetch_lengths.");
	    dbms_query_destroy(q); 
	    return -1;
	}

	/*Malloc a new structure */
	pnew_link_item = apr_pcalloc(db_r->p, sizeof(*pnew_link_item));
	presult_link_tail->next = pnew_link_item;
	presult_link_tail = pnew_link_item;

	/*Get the data of every column of this row to a link */

	presult_link_tail->serialno = atol(dbrow[0]);

        /* For Catacomb-V
	presult_link_tail->filename =
	    (dbrow[1] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[1],
						     lengths[1]);
	*/

	presult_link_tail->uri =
	    (dbrow[1] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[1], lengths[1]);
	DBG1("CKNIGHT DEBUG: %s", presult_link_tail->uri);

	presult_link_tail->creationdate = atoll(dbrow[2]);

	presult_link_tail->displayname =
	    (dbrow[3] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[3],
						     lengths[3]);

	presult_link_tail->getcontentlanguage =
	    (dbrow[4] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[4],
						     lengths[4]);

	presult_link_tail->getcontentlength = atol(dbrow[5]);
	presult_link_tail->getcontenttype = (dbrow[6] == NULL) ? NULL :
	    apr_pstrndup(db_r->p, dbrow[6], lengths[6]);

	presult_link_tail->getetag =
	    (dbrow[7] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[7],
						     lengths[7]);

	presult_link_tail->getlastmodified = atoll(dbrow[8]);

	presult_link_tail->resourcetype = atoi(dbrow[9]);

	presult_link_tail->source = atoi(dbrow[10]);

	presult_link_tail->depth = atoi(dbrow[11]);

	presult_link_tail->istext = atoi(dbrow[12]);
	
	presult_link_tail->isversioned = atoi(dbrow[13]);
	presult_link_tail->checkin = atoi(dbrow[14]);
	presult_link_tail->checkout = atoi(dbrow[15]);
	presult_link_tail->isexternal = atoi(dbrow[16]);

	presult_link_tail->next = NULL;
	presult_link_tail->pr = NULL;
    }				/*End of while (( dbrow = ... )) */

    ierrno = mysql_errno((MYSQL *) & (d->db));
    if (ierrno != 0) {		/* mysql_fetch_row() failed due to an error */
	db_error_message(db_r->p, &(d->db), "Failed when fetching");
	iret = -1;
    }
    else
	iret = 0;

    dbms_query_destroy(q); 

    return iret;
}				/*End of dbms_get_collection_resource */

/* Deletes a dasl_property by serialno and property name
 * @param d DB connection struct
 * @param r A pointer to a resource struct identifying 
 * the resource to be operated upon
 * @param pr A pointer to a property struct identifying the property to delete
 * @return 0 on success, 1 on failure
 * */
int dbms_del_dead_property(const dav_repos_db * d,
			   const dav_repos_resource * r,
			   const dav_repos_property * pr)
{
    apr_pool_t *pool = r->p;
    dav_repos_query *q = NULL;

    TRACE();

    q = dbms_prepare(pool, &(d->db), "Delete FROM dasl_property \
			       WHERE serialno=? and name=? and ns_id=?");
    dbms_set_int(q, 1, r->serialno);
    dbms_set_string(q, 2, pr->name);
    dbms_set_int(q, 3, pr->ns_id);

    if (dbms_execute(q)) {
	db_error_message(r->p, &(d->db), "Error in the delete command");
	dbms_query_destroy(q); 
	return -1;
    }

    dbms_query_destroy(q); 

    return 0;
}				/*End of dbms_del_dead_property */

/* Get resourcetype by uri
 * @param d DB connection struct
 * Return 0: success; 100: not found; -1: failure
 * */
int dbms_get_resourcetype(const dav_repos_db * d, dav_repos_resource * r)
{
    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    int iret = 0;
    MYSQL_ROW dbrow;
    apr_pool_t *pool = r->p;
    dav_repos_query *q = NULL;

    TRACE();

    q = dbms_prepare(pool, &(d->db), "SELECT resourcetype FROM dasl_resource \
			Where uri=?");
    dbms_set_string(q, 1, r->uri);

    if (!(use_result = dbms_select(q))) {
	db_error_message(r->p, &(d->db), "mysql_use_result error");
	dbms_query_destroy(q);
	return -1;
    }

    dbrow = mysql_fetch_row(use_result);	/*fetch one row */

    ierrno = mysql_errno((MYSQL *) (&(d->db)));
    if (ierrno != 0) {
	/* mysql_fetch_row() failed due to an error */
	db_error_message(r->p, &(d->db), "Failed when fetching");
	iret = -1;
    }
    else if (dbrow == NULL) {	/*No result */
	iret = 100;
    }
    else {
	/*One row fetched */
	/*Get the resourcetype data */
	/*Fetch value column of the row */
	r->resourcetype = atoi(dbrow[0]);

	iret = 0;
    }				/*End of else one row fetched */

    dbms_query_destroy(q); 

    return iret;
}				/*End of dbms_get_resourcetype */

/* Build namespace
 * 
 * Read namespace infor from DB and fill hash
 * db_r must have pool
 * This hash is for PROPFIND
 * */
void dbms_build_ns_id_hash(const dav_repos_db * db, dav_repos_resource * db_r)
{
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();
    
    /* id_hash is not clean */
    if (db_r->ns_id_hash != NULL)
	return;

    /* Build hash */
    DBG0("NS_id_hash_building");
    db_r->ns_id_hash = apr_hash_make(pool);
    
    q = dbms_prepare(pool, &(db->db), "select ns_id , name from dasl_namespace");
    
    /* Get ns_id */
    if (dbms_execute(q)) {
	db_error_message(pool, &(db->db), "mysql_use_result error");
        dbms_query_destroy(q);
	return;
    }
    /* Get all records */
    while (dbms_next(q)==1) {
	char *ns_id = dbms_get_string(q, 1);
	char *name = dbms_get_string(q, 2);
	
	if(ns_id && name) {
	    apr_hash_set(db_r->ns_id_hash, ns_id, APR_HASH_KEY_STRING, name);
	    DBG2("ADD Hash : %s: %s", ns_id, name);
	}
    }
    
    dbms_query_destroy(q);

    return;
}

/* Get namespace
 * If new namespace, save it and return ns_id
 * 
 * This hash is for PROPPATCH
 * It must be rebuild after adding new namespace
 * */
static void dbms_build_ns_hash(const dav_repos_db * db,
			       dav_repos_resource * db_r)
{
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();

    /* Already built */
    if (db_r->ns_hash)
	return;

    DBG0("NS_hash_building");
    db_r->ns_hash = apr_hash_make(db_r->p);

    
    q = dbms_prepare(pool, &(db->db), "select ns_id , name from dasl_namespace");
    
    /* Get ns_id */
    if (dbms_execute(q)) {
	db_error_message(pool, &(db->db), "mysql_use_result error");
        dbms_query_destroy(q);
	return;
    }
    
    /* Get all records */
    while (dbms_next(q)==1) {
	char *ns_id = dbms_get_string(q, 1);
	char *name = dbms_get_string(q, 2);
	
	if(ns_id && name) {
	    apr_hash_set(db_r->ns_hash, name, APR_HASH_KEY_STRING, ns_id);
	    DBG2("ADD Hash : %s: %s", ns_id, name);
	}
    }
    
    dbms_query_destroy(q);
    
    return;
}


/* Insert new namespace
 * Must build hash and try to find namespace first.
 */
static long dbms_insert_ns(const dav_repos_db * d,
			   dav_repos_resource * db_r, const char *namespace)
{
    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    MYSQL_ROW dbrow;
    long id;
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();

    /* We need to add name space to DB and fill hash */
    q = dbms_prepare(pool, &(d->db),
		     "insert into dasl_namespace (name) VALUES (?)");
    dbms_set_string(q, 1, namespace);

    if (dbms_execute(q)) {
	db_error_message(db_r->p, &(d->db), "mysql_query error");
        dbms_query_destroy(q); 
	return 0;
    }
    dbms_query_destroy(q);

    /* Success to insert */
    /* Get ns_id from DB */
    q = dbms_prepare(pool, &(d->db), "select * from dasl_namespace where name=?");
    dbms_set_string(q, 1, namespace);

    /* Get ns_id */
    if (!(use_result = dbms_select(q))) {
	db_error_message(db_r->p, &(d->db), "mysql_use_result error");
        dbms_query_destroy(q);
	return 0;
    }

    /* Get first result */
    dbrow = mysql_fetch_row(use_result);
    ierrno = mysql_errno((MYSQL *) (&(d->db)));
    if (ierrno != 0) {
	/* mysql_fetch_row() failed due to an error */
	db_error_message(db_r->p, &(d->db), "Failed when fetching");
        dbms_query_destroy(q);
	return 0;
    }


    /* Add to hash_table */
    if (db_r->ns_hash)
	apr_hash_set(db_r->ns_hash, apr_pstrdup(db_r->p, dbrow[1]),
		     APR_HASH_KEY_STRING, apr_pstrdup(db_r->p, dbrow[0]));

    id = atol(dbrow[0]);
    dbms_query_destroy(q); 

    return id;
}

/*## FIXME : Make it building hash only once per request
 *
 */
/* Get namespace
 * If new namespace, save it and return ns_id
 * 
 * @param d DB connection struct
 * @param r Pointer to a resource struct identifying the resource in question
 * @param namespace namespace
 * @return name space. Should not return NULL
 * */
const char *dbms_get_ns(const dav_repos_db * d,
			dav_repos_resource * db_r, const long ns_id)
{
    const char *namespace;
    const char *ns_id_str;
    const char *ns_default = "http://catacomb.tigris.org";

    TRACE();

    /* Build hash first */
    if (db_r->ns_id_hash == NULL)
	return ns_default;

    /* Build key */
    ns_id_str = apr_psprintf(db_r->p, "%ld", ns_id);
    namespace =
	(const char *) apr_hash_get(db_r->ns_id_hash, ns_id_str,
				    APR_HASH_KEY_STRING);

    /* We can't return NULL, so let's return default NS */
    if (namespace == NULL) {
	return ns_default;
    }

    return namespace;
}

/* Get namespace
 * If new namespace, save it and return ns_id
 * 
 * @param d DB connection struct
 * @param r Pointer to a resource struct identifying the resource in question
 * @param namespace namespace
 * @return ns_id
 * */
long dbms_get_ns_id(const dav_repos_db * d,
		    dav_repos_resource * db_r, const char *namespace)
{


    char *ns_id = NULL;

    TRACE();

    dbms_build_ns_hash(d, db_r);

    /* Now, We have Hash table */
    ns_id = apr_hash_get(db_r->ns_hash, namespace, APR_HASH_KEY_STRING);

    /* name space is not found */
    if (ns_id)
	return atol(ns_id);

    else
	return dbms_insert_ns(d, db_r, namespace);
}

/* This function for walker */
dav_error *dbms_load_locknull_list(const dav_repos_db * db,
				   char *path,
				   apr_text_header * phdr, apr_pool_t * pool)
{
    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    MYSQL_ROW dbrow;
    dav_repos_query *q = NULL;

    TRACE();
    dav_repos_no_trail(path);
    DBG0(path);

    q = dbms_prepare(pool, &(db->db), 
		     "select fname from dasl_locknull where path=?");
    dbms_set_string(q, 1, path);

    if (!(use_result = dbms_select(q))) {
	db_error_message(pool, &(db->db), "mysql_use_result error");
        dbms_query_destroy(q);
	return dav_new_error(pool,
			     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }

    while ((dbrow = mysql_fetch_row(use_result))) {
	ierrno = mysql_errno((MYSQL *) (&(db->db)));
	if (ierrno != 0) {
	    /* mysql_fetch_row() failed due to an error */
	    db_error_message(pool, &(db->db), "Failed when fetching");
            dbms_query_destroy(q);
	    return dav_new_error(pool,
				 HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
	}

	/* Add it to the text */
	apr_text_append(pool, phdr, apr_pstrdup(pool, dbrow[0]));
    }

    dbms_query_destroy(q);
    return NULL;
}

dav_error *dbms_save_locknull_member(const dav_repos_db * db,
				     const char *filename, apr_pool_t * pool)
{
    dav_repos_query *q = NULL;
    const char *fname = filename;
    char *path = ap_make_dirstr_parent(pool, filename);

    TRACE();

    /* no trailing */
    dav_repos_no_trail(path);

    /* We need to add name space to DB and fill hash */
    q = dbms_prepare(pool, &(db->db), "insert into dasl_locknull VALUES (?, ?)");
    dbms_set_string(q, 1, path);
    dbms_set_string(q, 2, fname);

    if (dbms_execute(q)) {
	db_error_message(pool, &(db->db), "mysql_query error");
        dbms_query_destroy(q);
	return dav_new_error(pool,
			     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }
    dbms_query_destroy(q);

    return NULL;
}

dav_error *dbms_remove_locknull_member(const dav_repos_db * db,
				       const char *filename,
				       apr_pool_t * pool)
{
    dav_repos_query *q = NULL;
    const char *fname = filename;
    char *path = ap_make_dirstr_parent(pool, filename);

    TRACE();

    /* no trailing */
    dav_repos_no_trail(path);

    /* We need to add name space to DB and fill hash */
    q = dbms_prepare(pool, &(db->db),
    	"DELETE from dasl_locknull where path=? and fname=?");
    dbms_set_string(q, 1, path);
    dbms_set_string(q, 2, fname);

    if (dbms_execute(q)) {
	db_error_message(pool, &(db->db), "mysql_query error");
	dbms_query_destroy(q);
	return dav_new_error(pool,
			     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }
    dbms_query_destroy(q);

    return NULL;
}


dav_error *dbms_load_lock_record(const dav_repos_db * db,
				 char *key,
				 dav_lock_discovery ** direct,
				 dav_lock_indirect ** indirect,
				 int *need_save,
				 apr_pool_t * pool)
{
    dav_lock_discovery *dp = NULL;
    dav_lock_indirect *ip = NULL;
    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    MYSQL_ROW dbrow;
    dav_repos_query *q = NULL;


    TRACE();
    dav_repos_no_trail(key);

    q = dbms_prepare(pool, &(db->db), "select locktype, scope, depth, "
		     "timeout, locktoken, owner, auth_user, lockkey "
		     "from dasl_lock where URI=?");
    dbms_set_string(q, 1, key);

    /* Get Results */
    if (!(use_result = dbms_select(q))) {
	db_error_message(pool, &(db->db), "mysql_use_result error");
        dbms_query_destroy(q);
	return dav_new_error(pool,
			     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }

    while ((dbrow = mysql_fetch_row(use_result))) {
	dav_locktoken *locktoken;

	ierrno = mysql_errno((MYSQL *) (&(db->db)));
	if (ierrno != 0) {
	    /* mysql_fetch_row() failed due to an error */
	    db_error_message(pool, &(db->db), "Failed when fetching");
	    dbms_query_destroy(q);
	    return dav_new_error(pool,
				 HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
	}

	/* Check timeout */
	if (dav_repos_lock_expired(atoll(dbrow[3]))) {
	    *need_save = DAV_TRUE;
	    continue;
	}

	locktoken = apr_pcalloc(pool, sizeof(*locktoken));
	locktoken->char_uuid = apr_pstrdup(pool, dbrow[4]);

	if (atoi(dbrow[0]) == DAV_LOCK_DIRECT) {

	    dp = apr_pcalloc(pool, sizeof(*dp));

	    /* Fill DP */
	    dp->f.scope = atoi(dbrow[1]);
	    dp->f.depth = atoi(dbrow[2]);
	    dp->f.timeout = atoll(dbrow[3]);

	    dp->locktoken = locktoken;

	    dp->owner = apr_pstrdup(pool, dbrow[5]);
	    if (dbrow[7] && strlen(dbrow[6]))
		dp->auth_user = apr_pstrdup(pool, dbrow[6]);

	    dp->next = *direct;
	    *direct = dp;
	}
	else {
	    ip = apr_pcalloc(pool, sizeof(*ip));

	    /* Fill IP */
	    ip->timeout = atoll(dbrow[3]);
	    ip->locktoken = locktoken;
	    ip->key = apr_pstrdup(pool, dbrow[7]);

	    /* Put it in the chain */
	    ip->next = *indirect;
	    *indirect = ip;
	}
    }

    dbms_query_destroy(q);

    return NULL;
}


/* DIRECT LOCK save
** lock info has linked list data
** #FIXME : Need rollback if we had error
*/
dav_error *dbms_save_lock_record(const dav_repos_db * db,
				 char *key,
				 const dav_lock_discovery * direct,
				 const dav_lock_indirect * indirect,
				 apr_pool_t * pool)
{
    dav_error *err;
    const dav_lock_discovery *dp;
    const dav_lock_indirect *ip;
    dav_repos_query *q = NULL;

    TRACE();
    dav_repos_no_trail(key);
    
    if (direct == NULL && indirect == NULL) { 
	DBG0("Nothing to save");
	return NULL;
    }
 
    /* FIXME: Should I remove all data?? 
    ** Should we find better way?
    */
    err = dbms_remove_lock_record(db, key ,pool);
    if( err != NULL) {
	return err;
    }
    
    for (dp = direct; dp; dp = dp->next) {
    	q = dbms_prepare(pool, &(db->db),
		"insert into dasl_lock (URI, locktype, scope, "
		"depth, timeout, locktoken, owner, auth_user) " 
		"VALUES (?, ?, ?, ?, ?, ?, ?, ?)");
	dbms_set_string(q, 1, key);
	dbms_set_int(q, 2, DAV_LOCK_DIRECT);
	dbms_set_int(q, 3, dp->f.scope);
	dbms_set_int(q, 4, dp->f.depth);
	dbms_set_int(q, 5, dp->f.timeout);
	dbms_set_string(q, 6, dp->locktoken->char_uuid);
	dbms_set_string(q, 7, dp->owner);
	dbms_set_string(q, 8, dp->auth_user);

	/* ## FIXME : Rollback?? */
	if (dbms_execute(q)) {
	    db_error_message(pool, &(db->db), "mysql_query error");
            dbms_query_destroy(q);
	    return dav_new_error(pool,
				 HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
	}
        dbms_query_destroy(q);
    }

    for (ip = indirect; ip; ip = ip->next) {
    	q = dbms_prepare(pool, &(db->db),
		"insert into dasl_lock \
			(URI, locktype, timeout, locktoken, lockkey) \
			VALUES (?, ?, ?, ?, ?)");
	dbms_set_string(q, 1, key);
	dbms_set_int(q, 2, DAV_LOCK_INDIRECT);
	dbms_set_int(q, 3, ip->timeout);
	dbms_set_string(q, 4, ip->locktoken->char_uuid);
	dbms_set_string(q, 5, ip->key);

	/* ## FIXME : Rollback?? */
	if (dbms_execute(q)) {
	    db_error_message(pool, &(db->db), "mysql_query error");
            dbms_query_destroy(q);
	    return dav_new_error(pool,
				 HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
	}
        dbms_query_destroy(q);
    }

    return NULL;
}

dav_error *dbms_remove_lock_record(const dav_repos_db * db,
				   char *key, apr_pool_t * pool)
{
    dav_repos_query *q = NULL;

    TRACE();
    dav_repos_no_trail(key);

    /* We need to add name space to DB and fill hash */
    q = dbms_prepare(pool, &(db->db), "DELETE from dasl_lock where URI=?");
    dbms_set_string(q, 1, key);

    if (dbms_execute(q)) {
	db_error_message(pool, &(db->db), "mysql_query error");
        dbms_query_destroy(q);
	return dav_new_error(pool,
			     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }
    dbms_query_destroy(q);

    return NULL;
}


/*
** FIXME : Can we just count without get the result??
*/
int dbms_lock_exists(const dav_repos_db * db,
		     char *key, apr_pool_t * pool)
{

    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    int iret = 0;
    MYSQL_ROW dbrow;
    dav_repos_query *q = NULL;

    TRACE();
    dav_repos_no_trail(key);

    /*Create the search command */
    q = dbms_prepare(pool, &(db->db), "SELECT URI From dasl_lock WHERE uri=?");
    dbms_set_string(q, 1, key);

    if (!(use_result = dbms_select(q))) {
	db_error_message(pool, &(db->db), "mysql_use_result error");
        dbms_query_destroy(q);
	return -1;
    }

    dbrow = mysql_fetch_row(use_result);	/*fetch one row */

    ierrno = mysql_errno((MYSQL *) (&(db->db)));
    if (ierrno != 0) {		/* mysql_fetch_row() failed due to an error */
	db_error_message(pool, &(db->db), "Failed when fetching");
	iret = -1;
    }
    else if (dbrow == NULL)	/*No result */
	iret = 0;
    else			/*One row fetched */
	iret = 1;

    dbms_query_destroy(q);

    return iret;
}				/*End of dbms_exists */

/**
 * Trun on the versioning for a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param db_r Identifies the resource to set the property on r->uri. 
 *          Contains the live property and value to set.
 * @checkin: Checkin version
 * @return NULL, for success; Error, otherwise.
 */
dav_error * dbms_set_is_versioned(const dav_repos_db * db, dav_repos_resource * db_r,
                       int checkin)
{
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();

    q = dbms_prepare(pool, &(db->db), "UPDATE dasl_resource set isversioned=1,"
                                      "checkin=? WHERE serialno=?");
    dbms_set_int(q, 1, checkin);
    dbms_set_int(q, 2, db_r->serialno);

    if (dbms_execute(q)) {
	db_error_message(pool, &(db->db), "mysql_query error");
        dbms_query_destroy(q);
	return dav_new_error(pool,
			     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }
    dbms_query_destroy(q);

    return NULL;
}				/*End of dbms_set_is_versioned*/

/**
 * Set the checkin and checkout version for a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param db_r Identifies the resource to set the property on r->uri. 
 *          Contains the live property and value to set.
 * @checkin: Checkin version
 * @checkout: Checkout version
 * @return NULL, for success; Error, otherwise.
 */
dav_error * dbms_set_checkin_out(const dav_repos_db * db, dav_repos_resource * db_r,
                       int checkin, int checkout)
{
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();

    q = dbms_prepare(pool, &(db->db), "UPDATE dasl_resource set checkin=?,"
                                      "checkout=? WHERE serialno=?");
    dbms_set_int(q, 1, checkin);
    dbms_set_int(q, 2, checkout);
    dbms_set_int(q, 3, db_r->serialno);

    if (dbms_execute(q)) {
	db_error_message(pool, &(db->db), "mysql_query error");
        dbms_query_destroy(q);
	return dav_new_error(pool,
			     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }
    dbms_query_destroy(q);

    return NULL;
}

/**
 * Set the checkin and checkout version for a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param db_r Identifies the resource to set the property on r->uri. 
 *          Contains the live property and value to set.
 * @author: The author who changed the resource
 * @return NULL, for success; Error, otherwise.
 */
dav_error * dbms_create_vr(const dav_repos_db * db, dav_repos_resource * db_r,
			   int version)
{
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;

    TRACE();

    /* Drop the temporary table first */
    q = dbms_prepare(pool, &(db->db), "drop table tmp_resforvcr");
    if (dbms_execute(q)) {
	DBG0("Error in the drop table command ");
    }
    dbms_query_destroy(q);

    /*Copy the record in the 'dasl_resource' table for 'version_resource' table */
    
    /*Store the source record into a temporary table */
    q = dbms_prepare(pool, &(db->db),
 			"CREATE temporary table tmp_resforvcr "
			"SELECT serialno, "
			"?+0+0," 
			"URI,"
			"?,"
			"displayname,"
			"getcontentlanguage,"
			"getcontentlength,"
			"getcontenttype,"
			"null," /* For getetag */
			"?+0,"
		        "resourcetype,"
			"source,"
			"depth,"
			"istext, "
			"isexternal, "
			"textcontent, "
			"bincontent, "
			"'author' "
			"FROM dasl_resource "
			"WHERE serialno=?");
    dbms_set_int(q, 1, version);
    dbms_set_int(q, 2, apr_time_now());
    dbms_set_int(q, 3, apr_time_now());
    dbms_set_int(q, 4, db_r->serialno);

    if (dbms_execute(q)) {
        db_error_message(pool, &(db->db),
	     "Error in the create temporary command");
        dbms_query_destroy(q); 
        return dav_new_error(pool,
		     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }
    dbms_query_destroy(q);

    /* Duplicate the record in the temporary table 
	   to the table version_resource */
    q = dbms_prepare(pool, &(db->db), "INSERT into version_resource "
		"SELECT * from tmp_resforvcr");
    if (dbms_execute(q)) {
        db_error_message(pool, &(db->db), "Error in the insert command");
        dbms_query_destroy(q); 
        return dav_new_error(pool,
    	     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }

    if (db_r->isexternal!=0) { // create a version of the file
	char *src_fn = NULL, *des_fn = NULL;
    	if (generate_path(&src_fn, db_r->p, db->file_dir,
		db_r->serialno, 0))
	    return dav_new_error(pool,
		HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to generate src fn");

    	if (generate_path(&des_fn, db_r->p, db->file_dir,
		db_r->serialno, version))
	    return dav_new_error(pool,
		HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to generate src fn");

	if(version != 0){
	  if (apr_file_copy(src_fn, des_fn, APR_OS_DEFAULT, db_r->p)
		!= APR_SUCCESS)
	    return dav_new_error(pool,
		HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to copy file.");
        }
    }

    dbms_query_destroy(q); 
    
    /*Copy the record in the 'dasl_property' table for 'version_property' table */
    
    /* Drop the temporary table first */
    q = dbms_prepare(pool, &(db->db), "drop table tmp_propertyforvcr");
    if (dbms_execute(q)) {
	DBG0("Error in the drop table command ");
    }
    dbms_query_destroy(q);

    /*Store the property record into a temporary table */
    q = dbms_prepare(pool, &(db->db),
 			"CREATE temporary table tmp_propertyforvcr "
			"SELECT dasl_property.serialno, "
			"?,"
			"ns_id,"
			"name,"
			"value "
			"FROM dasl_property,dasl_resource "
			"WHERE dasl_property.serialno=? AND "
			"dasl_property.serialno=dasl_resource.serialno");
    dbms_set_int(q, 1, version);
    dbms_set_int(q, 2, db_r->serialno);

    if (dbms_execute(q)) {
        db_error_message(pool, &(db->db),
	     "Error in the create temporary command");
        dbms_query_destroy(q); 
        return dav_new_error(pool,
		     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }
    dbms_query_destroy(q);

    /* Duplicate the records in the temporary table 
	   to the table version_property */
    q = dbms_prepare(pool, &(db->db), "INSERT into version_property "
		"SELECT * from tmp_propertyforvcr");
    if (dbms_execute(q)) {
        db_error_message(pool, &(db->db), "Error in the insert command");
        dbms_query_destroy(q); 
        return dav_new_error(pool,
    	     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }

    dbms_query_destroy(q); 

    return NULL;
}

/**
 * Find the versions of a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param db_r Identifies the resource to set the property on r->uri. 
 *          Contains the live property and value to set.
 * @vcrs: Versioned resource link
 * @return NULL, for success; Error, otherwise.
 */
dav_error * dbms_get_vrs(const dav_repos_db * db, dav_repos_resource * db_r,
                       dav_repos_resource ** vrs)
{
    MYSQL_RES *use_result = NULL;
    int ierrno = 0;
    MYSQL_ROW dbrow;
    unsigned long *lengths = NULL;
    struct dav_repos_resource *presult_link_tail = NULL;
    struct dav_repos_resource *pnew_link_item = NULL;
    apr_pool_t *pool = db_r->p;
    dav_repos_query *q = NULL;
    
    TRACE();
    
    *vrs = NULL;

    /*Create the search command */
    q = dbms_prepare(pool, &(db->db), "SELECT serialno,"
                "version,"
		"URI,"
		"creationdate,"
		"displayname,"
		"getcontentlanguage,"
		"getcontentlength,"
		"getcontenttype,"
		"getetag,"
		"getlastmodified,"
		"resourcetype,"
		"source,"
		"depth, "
		"istext "
		"From version_resource Where serialno = ? "
		"ORDER by version DESC");
    dbms_set_int(q, 1, db_r->serialno);

    if (!(use_result = dbms_select(q))) {
	db_error_message(db_r->p, &(db->db), "mysql_use_result error");
	dbms_query_destroy(q);
        return dav_new_error(pool,
		     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }

    while ((dbrow = mysql_fetch_row(use_result))) {
	/*fetch every row */
	/*Get the length of every field */
	lengths = mysql_fetch_lengths(use_result);
	if (lengths == NULL) {
	    db_error_message(db_r->p, &(db->db), "Error in mysql_fetch_lengths.");
	    dbms_query_destroy(q); 
            return dav_new_error(pool,
		     HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
	}

	/*Malloc a new structure */
	pnew_link_item = apr_pcalloc(db_r->p, sizeof(*pnew_link_item));
	if (presult_link_tail == NULL) { /* The first one */
	    *vrs = pnew_link_item;
	    presult_link_tail = pnew_link_item;
	    presult_link_tail->next = NULL;
	}
	else { /* Not the first one */
	    presult_link_tail->next = pnew_link_item;
	    presult_link_tail = pnew_link_item;
        }

	/*Get the data of every column of this row to a link */

	presult_link_tail->serialno = atol(dbrow[0]);

	presult_link_tail->version = atol(dbrow[1]);

	presult_link_tail->uri =
	    (dbrow[2] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[2], lengths[2]);
	DBG1("CKNIGHT DEBUG: %s", presult_link_tail->uri);

	presult_link_tail->creationdate = atoll(dbrow[3]);

	presult_link_tail->displayname =
	    (dbrow[4] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[4],
						     lengths[4]);

	presult_link_tail->getcontentlanguage =
	    (dbrow[5] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[5],
						     lengths[5]);

	presult_link_tail->getcontentlength = atol(dbrow[6]);
	presult_link_tail->getcontenttype = (dbrow[7] == NULL) ? NULL :
	    apr_pstrndup(db_r->p, dbrow[7], lengths[7]);

	presult_link_tail->getetag =
	    (dbrow[8] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[8],
						     lengths[8]);

	presult_link_tail->getlastmodified = atoll(dbrow[9]);

	presult_link_tail->resourcetype = atoi(dbrow[10]);

	presult_link_tail->source = atoi(dbrow[11]);

	presult_link_tail->depth = atoi(dbrow[12]);

	presult_link_tail->istext = atoi(dbrow[13]);
	
	presult_link_tail->next = NULL;
	presult_link_tail->pr = NULL;
    }				/*End of while (( dbrow = ... )) */

    ierrno = mysql_errno((MYSQL *) & (db->db));
    if (ierrno != 0) {		/* mysql_fetch_row() failed due to an error */
	db_error_message(db_r->p, &(db->db), "Failed when fetching");
        dbms_query_destroy(q); 
        return dav_new_error(pool,
	         HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error");
    }

    dbms_query_destroy(q); 

    return NULL;
}

struct dbms_stream {
    dav_repos_db *db;
    dav_repos_resource *db_r;
    apr_file_t *file;
    char *path;
    int temporary;
};

dav_error *dbms_open_stream(dav_repos_db *db, dav_repos_resource *db_r,
	dbms_stream **stream, dav_stream_mode mode) {
    dav_error *err = NULL;
    dbms_stream *s = NULL;
    dav_repos_query *q = NULL;

    TRACE();

    s = apr_pcalloc(db_r->p, sizeof(*s));
    s->db = db;
    s->db_r = db_r;

    // only way to get a serialno is to add a row (or convert to using a
    // sequence-generator
    if (!db_r->serialno) {
	long serialno;
	if (db_insert_resource(db, db_r, &serialno))
	    return dav_new_error(db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
		"Unable to add resource to the database");
	db_r->serialno = serialno;
    }

    DBG1("Max_Filesize: %li", db->max_file_size );
    DBG1("Actual_contentlength: %li", db_r->getcontentlength );

    // save only in filesystem if max_file_size is smaller than contentlength
    if ( (db->file_dir != NULL) && (db->max_file_size < db_r->getcontentlength) ) { 
    	if ((err = generate_path(&(s->path), db_r->p, db->file_dir,
		db_r->serialno, db_r->version))) {
	    DBG0("error in generating path, removing resource record.");
	    q = dbms_prepare(db_r->p, &(db->db), "DELETE FROM dasl_resource "
		"WHERE serialno=?");
	    dbms_set_int(q, 1, db_r->serialno);
	    dbms_execute(q);
	    dbms_query_destroy(q);
	}
    }

    if (!s->path) {
	s->temporary = 1;
	s->path = apr_psprintf(db_r->p, "%s/catacomb-%ld-%d",
		db->tmp_dir, db_r->serialno, db_r->version);
	ap_no2slash(s->path);
    }

    DBG1("path: %s", s->path);

    switch(mode) {
    	case DAV_MODE_WRITE_TRUNC:
    	case DAV_MODE_WRITE_SEEKABLE:
	    if (apr_file_open(&(s->file), s->path,
		    APR_CREATE | APR_WRITE | APR_TRUNCATE | APR_BINARY | APR_BUFFERED,
		    APR_OS_DEFAULT, db_r->p) != APR_SUCCESS)
		err = dav_new_error(db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
		    "Unable to open file for write");
	    break;
    	default: // any possibility of this happening?
	    if (apr_file_open(&(s->file), s->path,
		    APR_READ | APR_BINARY | APR_BUFFERED,
		    APR_OS_DEFAULT, db_r->p) != APR_SUCCESS)
		err = dav_new_error(db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
		    "Unable to open file for read");
	    break;
    }

    if (err) {
	    DBG0("error in opening file");
	    q = dbms_prepare(db_r->p, &(db->db), "DELETE FROM dasl_resource "
		"WHERE serialno=?");
	    dbms_set_int(q, 1, db_r->serialno);
	    dbms_execute(q);
	    dbms_query_destroy(q);
	    return err;
    }

    *stream = s;
    return NULL;
}

dav_error *dbms_close_stream(dbms_stream *stream, int commit) {
    // TODO: merge dbms_write_content into this
    dav_error *err = NULL;
    dav_repos_query *q = NULL;

    TRACE();

    if (commit) {
    	if (apr_file_close(stream->file) == APR_SUCCESS) {
	    if (stream->temporary) {
		if (dbms_write_content(stream->db, stream->db_r,
			stream->path)) {
		    err = dav_new_error(stream->db_r->p,
			HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to close file!");
		}
	    }
	} else err = dav_new_error(stream->db_r->p,
	    HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to close file!");
    }

    if (!commit) {
	q = dbms_prepare(stream->db_r->p, &(stream->db->db), "DELETE FROM dasl_resource "
	    "WHERE serialno=?");
	dbms_set_int(q, 1, stream->db_r->serialno);
	dbms_execute(q);
	dbms_query_destroy(q);
    }
    
    if (stream->temporary) {
	if (apr_file_remove(stream->path, stream->db_r->p) != APR_SUCCESS)
	    err = dav_new_error(stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR,
	        0, "Unable to remove temporary file.");
    }
    return err;
}

dav_error *dbms_write_stream(dbms_stream *stream, const void *buf,
	const apr_size_t bufsize) {
    apr_size_t s = bufsize;
    TRACE();

    DBG1("stream = %p", stream);
    DBG1("stream->file = %p", stream->file);
    DBG1("contentlength_of_file = %li", stream->db_r->getcontentlength);
    if (apr_file_write(stream->file, buf, &s) != APR_SUCCESS)
        return dav_new_error(stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
	    "Unable to write to file.");
    if (s != bufsize)
        return dav_new_error(stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
	    "Did not write all contents.");
    return NULL;
}

dav_error *dbms_seek_stream(dbms_stream *stream, const apr_size_t pos) {
    apr_size_t p = pos;
    TRACE();

    if (apr_file_write(stream->file, APR_SET, &p) != APR_SUCCESS)
        return dav_new_error(stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
	    "Unable to seek in file.");
    if (p != pos)
        return dav_new_error(stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
	    "Seek resulted in different position.");
    return NULL;
}

dav_error *dbms_deliver(dav_repos_db *db, dav_repos_resource *db_r,
	ap_filter_t *output) {
    dav_error *err = NULL;
    char *filename = NULL;
    apr_file_t *fd;
    apr_bucket_brigade *bb;
    apr_bucket *bkt;

    if ( (db->file_dir) && (db->max_file_size < db_r->getcontentlength) ) {
	err = generate_path(&filename, db_r->p, db->file_dir,db_r->serialno, db_r->version);
        if (err) 
	    return err;
    } else {
	/* Anyway we need to make random filename */
	filename = apr_psprintf(db_r->p, "%s/dav_repos_deliver_XXXXXX",
	    db->tmp_dir);
	if (mktemp(filename) == NULL) {
	    return dav_new_error(db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
		"An error occurred while opening a resource name.");
	}

	/* get file from data base */
	if (dbms_read_content(db, db_r, filename) != 0) {
	    return dav_new_error(db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
		"An error occurred while getting a resource name.");
	}
    }

DBG1("TESTCDK: %s", filename);
    if (apr_file_open(&fd, filename, APR_READ | APR_BINARY, 0,
	    db_r->p) != APR_SUCCESS) {
	return dav_new_error(db_r->p, HTTP_FORBIDDEN, 0,
			     "File permissions deny server access.");
    }

    bb = apr_brigade_create(db_r->p, output->c->bucket_alloc);

    /* ### this does not handle large files. but this is test code anyway */
    bkt = apr_bucket_file_create(fd, 0,
				 (apr_size_t) db_r->getcontentlength,
				 db_r->p, output->c->bucket_alloc);
    APR_BRIGADE_INSERT_TAIL(bb, bkt);

    bkt = apr_bucket_eos_create(output->c->bucket_alloc);
    APR_BRIGADE_INSERT_TAIL(bb, bkt);

    if (ap_pass_brigade(output, bb) != APR_SUCCESS)
	err = dav_new_error(db_r->p, HTTP_FORBIDDEN, 0,
			     "Could not write contents to filter.");

    if (!db->file_dir) apr_file_remove(filename, db_r->p); //ignore the error?
    return err;
}
