/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <httpd.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_core.h>          /* for ap_construct_url */
#include <mod_dav.h>
#include <mysql.h>
#include <mysqld_error.h>
#include <apr_strings.h>

#include "dav_repos.h"
#include "dbms_mysql.h"

/* utility function to print the error code */
static inline int debug_mysql(const int status, MYSQL *mysql,
	const char *db_error_message_str) {
    if (status) {
	DBG3("Error: %s\nError Number: %d\nErrorMessage:%s\n",
	    db_error_message_str, mysql_errno(mysql), mysql_error(mysql));
    }
    return status;
}

#ifdef DEBUG
#define DEBUG_APR_BUF_SIZE 160 /* are two lines sufficient? */
static char debug_apr_buf[DEBUG_APR_BUF_SIZE];
#endif
static inline int debug_apr(const apr_status_t status) {
    if(status)
	DBG0(apr_strerror(status, debug_apr_buf, DEBUG_APR_BUF_SIZE));
    return status;
}

#ifdef DEBUG
static int dbms_count = 0;
static int query_count = 0;
#endif

int dbms_create(dav_repos_dbms *db, const char *user,
	const char *password, const char *database, const char *host,
	const int port) {
    mysql_init(db);
    /* TODO: fix to support the UNIX domain socket option */
    mysql_real_connect(db, host, user, password, database, port, NULL,
	0);
#ifdef DEBUG
    dbms_count++;
    DBG1("DB CREATED: %d\n", dbms_count);
#endif
    return 0;
}

int dbms_destroy(dav_repos_dbms *db) {
    mysql_close(db);
#ifdef DEBUG
    dbms_count--;
    DBG1("DB DESTROYED: %d\n", dbms_count);
#endif
    return 0;
}

const char *dbms_error(apr_pool_t *pool, const dav_repos_dbms *db) {
    return mysql_error((MYSQL *)db);
}

int dbms_errno(const dav_repos_dbms *db) {
    return mysql_errno((MYSQL *)db);
}

/* TODO: should queries have their own (child) pool(s)? */
dav_repos_query *dbms_prepare(apr_pool_t *pool,
	const dav_repos_dbms *dbms, const char *query) {
    dav_repos_query *q;
    int i, query_len;

    query_len = strlen(query);

    q = (dav_repos_query *)apr_pcalloc(pool, sizeof(dav_repos_query));
    q->pool = pool;
    q->db = (dav_repos_dbms *)dbms;
    q->query_string = (char *)apr_pstrndup(pool, query, query_len);
    q->state = DAV_REPOS_STATE_PREPARED;
    for(i = 0, q->param_count = 0; i < query_len; i++) {
	if (query[i] == '?') q->param_count++;
    }
    /* should I use apr_pcalloc instead? */
    q->parameters = (char **)apr_pcalloc(pool, sizeof(char *) * q->param_count);
    for (i = 0; i < q->param_count; i++) q->parameters[i] = NULL;

#ifdef DEBUG
    query_count++;
    DBG1("QUERY CREATED: %d\n", query_count);
#endif
    return q;
}

#define CHECK_PARAMS if (num < 1 || num > query->param_count) return 0;
int dbms_set_int(dav_repos_query *query,
	const int num, const long long value) {
    CHECK_PARAMS
    query->parameters[num - 1] = apr_psprintf(query->pool, "%lld", value);
    return 0;
}

int dbms_set_float(dav_repos_query *query,
	const int num, const double value) {
    CHECK_PARAMS
    query->parameters[num - 1] = apr_psprintf(query->pool, "%f", value);
    return 0;
}

int dbms_set_string(dav_repos_query *query,
	const int num, const char *value) {
    long value_length, enc_length;
    char *param;

    CHECK_PARAMS
    if (!value) {
	query->parameters[num - 1] = 
	  apr_pstrdup(query->pool, "null");
	return 0;
    }
    
    value_length = strlen(value);
    param = query->parameters[num - 1] = apr_pcalloc(query->pool,
    	value_length * 2 + 3);
    enc_length = mysql_real_escape_string(query->db, param + 1, value,
    	value_length);
    *param = '\'';
    param[enc_length + 1] = '\'';
    param[enc_length + 2] = '\0';
    return 0;
}

int dbms_execute(dav_repos_query *query) {
    int full_length, query_string_length;
    int i, j, k, error;
    char *escquery;

    full_length = query_string_length = strlen(query->query_string);

    for (i = 0; i < query->param_count; i++) {
	if (!query->parameters[i]) return 0; /* parameter was never set */
	full_length += strlen(query->parameters[i]) - 1; /* dont count orig ? */
    }

    /* TODO: optimize to use pointers, not offsets */
    if (query->param_count==0) {
	escquery = apr_pstrdup(query->pool, query->query_string);
    } else {

	/* make space for the trailing '\0' */
	escquery = apr_pcalloc(query->pool, full_length + 1);

	for (i = 0, j = 0, k = 0; i < query_string_length; i++) {
	    
	    if (query->query_string[i] == '?') {
#if 0 
		if( k >= query->param_count) {
		    break;
#endif	
		strcpy(escquery + j, query->parameters[k]);
		j += strlen(query->parameters[k++]);
	    } else {
		escquery[j++] = query->query_string[i];
	    }
	}
	escquery[j] = 0;
    }
    DBG1("Query to execute: %s", escquery);
    //error = mysql_real_query(query->db, escquery, full_length);
    error = mysql_query(query->db, escquery);
    if (error) {
        query->state = DAV_REPOS_STATE_ERROR;
    	return error;
    }
    
    if (!strncasecmp("select", escquery, 6)) { 
	   query->results = mysql_use_result(query->db);
	   query->colcount = mysql_field_count(query->db);
       query->lengths = apr_pcalloc(query->pool,
	   sizeof(long) * query->colcount);
    }

    query->state = DAV_REPOS_STATE_EXECUTED;
    return 0;
}

MYSQL_RES *dbms_select(dav_repos_query *query) {
    DBG0("dbms_select depricated, do not use!\n");
    if (dbms_execute(query)) return NULL;
    return query->results;
}

int dbms_next(dav_repos_query *query) {
    int i;
    long *lengths;

    if (query->state != DAV_REPOS_STATE_EXECUTED) return -1;
    if (!query->results) query->results = mysql_use_result(query->db);
    query->row = mysql_fetch_row(query->results);
    if (!query->row) return 0;
    if (mysql_errno(query->db)) return -1;
    lengths = mysql_fetch_lengths(query->results);
    // Should I use strndup or some other fn to dup?
    for (i = 0; i < query->colcount; i++) query->lengths[i] = lengths[i];
    return 1;
}

long long dbms_get_int(dav_repos_query *query,
	int column) {
    if (!query->row) return 0; /* should this do something different? */
    if (column < 1 || column > query->colcount + 1) return 0;
    /* TODO: should check column type. */
    return atoll(query->row[column - 1]);
}

char *dbms_get_string(dav_repos_query *query,
	int column) {
    if (!query->row) return 0; /* should this do something different? */
    if (column < 1 || column > query->colcount + 1) return 0;
    /* TODO: should check column type. */
    if (query->row[column - 1] == NULL) return NULL;
    return apr_pstrndup(query->pool, query->row[column - 1],
        query->lengths[column - 1]);
}

int dbms_query_destroy(dav_repos_query *query) {
    if (query->results) {
	mysql_free_result(query->results);
	query->results = NULL;
    }
#ifdef DEBUG
    query_count--;
    DBG1("QUERY DESTROYED: %d\n", query_count);
#endif
    return 0;
}


