/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#ifndef DBMS_MYSQL_H
#define DBMS_MYSQL_H
#include <mysql.h>
#include <apr_pools.h>

/**
 * This piece of code provides a JDBC-like API for query execution. In
 * particular, it provides for parameter substitution (with parameter value
 * escaping.) This *may* actually improve performance if you have multiple
 * queries sharing some common parameters. An abstract example:
 *
 * dav_repos_dbms db;
 * dav_repos_query *q;
 * MYSQL_RES *res;
 * apr_pool_t *pool;
 * apr_pool_create(&pool, NULL);
 *
 * if (!dbms_create(pool, &db, "root", "root", "repos", null, 0)) exit(1);
 * q = dbms_prepare(pool, &db,
 *   "select address from employee where name = ? and officenum = ?");
 * dbms_set_string(pool, q, 1, "Chris Knight");
 * dbms_set_int(pool, q, 2, 295);
 * res = dbms_select(pool, q);
 * if (!res) { apr_pool_destroy(pool); exit(ERR); }
 * ... (use the result as you would normally)
 * dbms_set_string(pool, q, 2, 101);
 * res = dbms_select(pool, q);
 * if (!res) { apr_pool_destroy(pool); exit(ERR); }
 * ...
 * dbms_destroy(&db);
 * apr_pool_destroy(pool);
 * exit(0);
 */

typedef MYSQL dav_repos_dbms;

#define DAV_REPOS_TYPE_INT 1
#define DAV_REPOS_TYPE_FLOAT 2
#define DAV_REPOS_TYPE_STRING 3

#define DAV_REPOS_STATE_PREPARED 1
#define DAV_REPOS_STATE_EXECUTED 2
#define DAV_REPOS_STATE_ERROR 3

typedef struct {
    dav_repos_dbms *db;
    char *query_string;
    char **parameters;
    short int *parameter_type; /* one of DAV_REPOS_TYPE_* */
    int param_count;
    MYSQL_RES *results;
    long *lengths;
    int colcount;
    MYSQL_ROW row;
    short int state; /* one of DAV_REPOS_STATE_* */

    apr_pool_t *pool; /* Inner use */
} dav_repos_query;

/**
 * Initializes the database connection and memory pool. Returns true on
 * success.
 *
 * @see #dbms_destroy
 * @param host - The hostname to connect to. If null, then the UNIX-domain
 *   socket is used. (And port is useless.)
 * @param port - Specifies the port to connect to, or 0 if you would like
 *   to use the default port.
 * @return 0 on success
 */
int dbms_create(dav_repos_dbms *db,
    const char *user, const char *password, const char *database,
    const char *host, int port);

/**
 * Destroys the database connection for this database.
 * @return 0 on success
 */
int dbms_destroy(dav_repos_dbms *db);

/**
 * Generates an error message with database specifics. Remember, the error
 * message string may come from this pool, so do something with it before
 * destroying the pool.
 */
const char *dbms_error(apr_pool_t *pool, const dav_repos_dbms *db);

/**
 * Returns a database error number.
 * TODO: are there standard error number schemes?
 *   I assume 0 is "no error".
 */
int dbms_errno(const dav_repos_dbms *db);

/**
 * Creates a new dav_repos_query object. Returns null when there is a failure.
 *
 * @param pool - The memory pool to allocate from.
 * @param dbms - The database to perform the query on.
 * @param query - The query you would like to execute. Any parameters you will
 *   pass to this query should be denoted by "?" placeholders in the query
 *   string.  I.e. "select address from employee where name = ?"  Note you
 *   should *not* place quotes around placeholders you plan on replacing with
 *   strings.
 * @return 0 on success
 * @see #dbms_set_int
 */
dav_repos_query *dbms_prepare(apr_pool_t *pool,
	const dav_repos_dbms *dbms, const char *query);

/**
 * Replaces a placeholder "?" with an integer value.
 *
 * @param query - The query that you would like to place the value in.
 * @param num - The number of the "?" to replace (the first is #1, etc.)
 * @param value - The value to place at that position.
 * @return 0 on success
 * @see #dbms_set_float
 * @see #dbms_set_string
 */
int dbms_set_int(dav_repos_query *query,
	const int num, const long long value);

/**
 * Replaces a placeholder "?" with an floating-point value.
 *
 * @param query - The query that you would like to place the value in.
 * @param num - The number of the "?" to replace (the first is #1, etc.)
 * @param value - The value to place at that position.
 * @return 0 on success
 * @see #dbms_set_int
 * @see #dbms_set_string
 */
int dbms_set_float(dav_repos_query *query,
	const int num, const double value);

/**
 * Replaces a placeholder "?" with a string value. Note, this creates an
 * escaped version of this string in the dbms's memory pool.
 *
 * @param query - The query that you would like to place the value in.
 * @param num - The number of the "?" to replace (the first is #1, etc.)
 * @param value - The value to place at that position.
 * @return 0 on success
 * @see #dbms_set_int
 * @see #dbms_set_float
 */
int dbms_set_string(dav_repos_query *query,
    const int num, const char *value);

/**
 * Executes the query.
 *
 * @return 0 on success
 */
int dbms_execute(dav_repos_query *query);

/**
 * Executes the query.
 *
 * @return a pointer to a MYSQL_RES 
 * @depricated
 */
MYSQL_RES *dbms_select(dav_repos_query *query);

/**
 * Fetches the next row of results.
 *
 * @return 1 if success, 0 if no remaining rows, -1 if there is an error.
 */
int dbms_next(dav_repos_query *query);

/**
 * Retrieve an integer value from the database.
 *
 * @param column - The column number in the result set, starting with 1.
 * @return The value of the column in the current row.
 * @see #dbms_next
 * @see #get_string
 */
long long dbms_get_int(dav_repos_query *query, int column);

/**
 * Retrieve an string value from the database. Note that the memory buffer
 * allocated for this is taken from the pool, so be sure to duplicate it
 * before you destroy the pool!
 *
 * @param column - The column number in the result set, starting with 1.
 * @return The value of the column in the current row.
 * @see #dbms_next
 * @see #get_int
 */
char *dbms_get_string(dav_repos_query *query, int column);

/**
 * Releases any resources allocated for this query.  Currently does nothing.
 * 
 * @return 0 on success.
 */
int dbms_query_destroy(dav_repos_query *query);

#endif /*DBMS_MYSQL_H*/

